# Copyright 2019-, Gavin E. Crooks and contributors
# Copyright 2016-2018, Rigetti Computing
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.


from typing import TextIO

import numpy as np
import scipy

from . import tensors, utils
from .ops import Gate, Unitary
from .paulialgebra import Pauli
from .qubits import Qubit, Qubits
from .tensors import QubitTensor


# DOCME: Explain cv_interchangeable
class RandomGate(Unitary):
    r"""Returns a random unitary gate on K qubits.

    Ref:
        "How to generate random matrices from the classical compact groups"
        Francesco Mezzadri, math-ph/0609050
    """
    cv_interchangeable = True

    def __init__(self, qubits: Qubits) -> None:
        qubits = tuple(qubits)
        tensor = utils.unitary_ensemble(2 ** len(qubits))
        super().__init__(tensor, qubits)


# end class RandomGate


# DOCME: Can also use circuit.asgate()
def join_gates(gate0: Gate, gate1: Gate) -> Unitary:
    """Direct product of gates. Qubit count is the sum of each gate's
    bit count."""
    tensor = tensors.outer(gate0.tensor, gate1.tensor, rank=2)
    return Unitary(tensor, tuple(gate0.qubits) + tuple(gate1.qubits))


def print_gate(gate: Gate, ndigits: int = 2, file: TextIO = None) -> None:
    """Pretty print a gate tensor

    Args:
        gate:
        ndigits:
        file: Stream to which to write. Defaults to stdout
    """
    N = gate.qubit_nb
    gate_tensor = gate.tensor
    lines = []
    for index, amplitude in np.ndenumerate(gate_tensor):
        ket = "".join([str(n) for n in index[0:N]])
        bra = "".join([str(index[n]) for n in range(N, 2 * N)])
        if round(abs(amplitude) ** 2, ndigits) > 0.0:
            lines.append(f"{bra} -> {ket} : {amplitude}")
    lines.sort(key=lambda x: int(x[0:N]))
    print("\n".join(lines), file=file)


# FIXME: P0 and P1 should be channels or Projectors ???


class P0(Gate):
    r"""Project qubit to zero.

    A non-unitary gate that represents the effect of a measurement. The norm
    of the resultant state is multiplied by the probability of observing 0.
    """
    text_labels = ["|0><0|"]

    def __init__(self, q0: Qubit = 0) -> None:
        super().__init__(qubits=[q0])

    @utils.cached_property
    def tensor(self) -> QubitTensor:
        return tensors.asqutensor([[1, 0], [0, 0]])


class P1(Gate):
    r"""Project qubit to one.

    A non-unitary gate that represents the effect of a measurement. The norm
    of the resultant state is multiplied by the probability of observing 1.
    """
    text_labels = ["|1><1|"]

    def __init__(self, q0: Qubit = 0) -> None:
        super().__init__(qubits=[q0])

    @utils.cached_property
    def tensor(self) -> QubitTensor:
        return tensors.asqutensor([[0, 0], [0, 1]])


def unitary_from_hamiltonian(hamiltonian: Pauli, qubits: Qubits) -> Unitary:
    """Create a Unitary gate U from a Pauli operator H, U = exp(-i H)"""
    # Note: Can't be a classmethod constructor on Unitary due to circular
    # imports. (???)
    op = hamiltonian.asoperator(qubits)
    U = scipy.linalg.expm(-1j * op)
    return Unitary(U, qubits)


# TODO: Replace with class
def conditional_gate(control: Qubit, gate0: Gate, gate1: Gate) -> Unitary:
    """Return a conditional unitary gate. Do gate0 on bit 1 if bit 0 is zero,
    else do gate1 on 1"""
    assert gate0.qubits == gate1.qubits  # FIXME

    tensor = join_gates(P0(control), gate0).tensor
    tensor += join_gates(P1(control), gate1).tensor
    gate = Unitary(tensor, (control,) + tuple(gate0.qubits))
    return gate


# fin



# Copyright 2019-, Gavin E. Crooks and the QuantumFlow contributors
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

import urllib

import pytest

import quantumflow as qf
from quantumflow.xquirk import circuit_to_quirk, quirk_url


def test_circuit_to_quirk() -> None:
    # 2-qubit gates

    quirk = "https://algassert.com/quirk#circuit={%22cols%22:[[1,%22X%22,%22%E2%80%A2%22],[%22%E2%80%A2%22,1,%22Z%22],[1,%22%E2%80%A2%22,%22Y%22],[%22Swap%22,1,%22Swap%22]]}"  # noqa: E501
    circ = qf.Circuit([qf.CNot(2, 1), qf.CZ(0, 2), qf.CY(1, 2), qf.Swap(0, 2)])
    print()
    print(urllib.parse.unquote(quirk))
    print(quirk_url(circuit_to_quirk(circ)))
    assert urllib.parse.unquote(quirk) == quirk_url(circuit_to_quirk(circ))

    # 3-qubit gates
    quirk = "https://algassert.com/quirk#circuit={%22cols%22:[[%22%E2%80%A2%22,%22%E2%80%A2%22,%22X%22],[%22%E2%80%A2%22,%22%E2%80%A2%22,%22Z%22],[%22%E2%80%A2%22,%22Swap%22,%22Swap%22]]}"  # noqa: E501
    circ = qf.Circuit([qf.CCNot(0, 1, 2), qf.CCZ(0, 1, 2), qf.CSwap(0, 1, 2)])
    print()
    print(urllib.parse.unquote(quirk))
    print(quirk_url(circuit_to_quirk(circ)))
    assert urllib.parse.unquote(quirk) == quirk_url(circuit_to_quirk(circ))

    test0 = "https://algassert.com/quirk#circuit={%22cols%22:[[%22Z%22,%22Y%22,%22X%22,%22H%22]]}"  # noqa: E501
    test0 = urllib.parse.unquote(test0)
    circ = qf.Circuit([qf.Z(0), qf.Y(1), qf.X(2), qf.H(3)])
    print(test0)
    print(quirk_url(circuit_to_quirk(circ)))
    assert test0 == quirk_url(circuit_to_quirk(circ))

    test_halfturns = "https://algassert.com/quirk#circuit={%22cols%22:[[%22X^%C2%BD%22,%22Y^%C2%BD%22,%22Z^%C2%BD%22],[%22X^-%C2%BD%22,%22Y^-%C2%BD%22,%22Z^-%C2%BD%22]]}"  # noqa: E501
    test_halfturns = urllib.parse.unquote(test_halfturns)
    circ = qf.Circuit(
        [qf.V(0), qf.SqrtY(1), qf.S(2), qf.V(0).H, qf.SqrtY(1).H, qf.S(2).H]
    )
    print(test_halfturns)
    print(quirk_url(circuit_to_quirk(circ)))
    assert test_halfturns == quirk_url(circuit_to_quirk(circ))

    quarter_turns = "https://algassert.com/quirk#circuit={%22cols%22:[[%22Z^%C2%BC%22],[%22Z^-%C2%BC%22]]}"  # noqa: E501
    s = urllib.parse.unquote(quarter_turns)
    circ = qf.Circuit([qf.T(0), qf.T(0).H])
    assert s == quirk_url(circuit_to_quirk(circ))

    # GHZ circuit
    quirk = "https://algassert.com/quirk#circuit={%22cols%22:[[%22H%22],[%22%E2%80%A2%22,%22X%22],[1,%22%E2%80%A2%22,%22X%22]]}"  # noqa: E501
    circ = qf.Circuit([qf.H(0), qf.CNot(0, 1), qf.CNot(1, 2)])
    print(urllib.parse.unquote(quirk))
    print(quirk_url(circuit_to_quirk(circ)))
    assert urllib.parse.unquote(quirk) == quirk_url(circuit_to_quirk(circ))

    test_formulaic = "https://algassert.com/quirk#circuit={%22cols%22:[[{%22id%22:%22X^ft%22,%22arg%22:%220.1%22},{%22id%22:%22Y^ft%22,%22arg%22:%220.2%22},{%22id%22:%22Z^ft%22,%22arg%22:%220.3%22}],[{%22id%22:%22Rxft%22,%22arg%22:%220.4%22},{%22id%22:%22Ryft%22,%22arg%22:%220.5%22},{%22id%22:%22Rzft%22,%22arg%22:%220.6%22}]]}"  # noqa: E501
    s = urllib.parse.unquote(test_formulaic)
    circ = qf.Circuit(
        [
            qf.XPow(0.1, 0),
            qf.YPow(0.2, 1),
            qf.ZPow(0.3, 2),
            qf.Rx(0.4, 0),
            qf.Ry(0.5, 1),
            qf.Rz(0.6, 2),
        ]
    )
    assert s == quirk_url(circuit_to_quirk(circ))


def test_fail() -> None:
    with pytest.raises(ValueError):
        circ = qf.Circuit([qf.Can(0.1, 0.2, 0.3, 0, 1)])
        quirk_url(circuit_to_quirk(circ))


def test_url_escape() -> None:
    circ = qf.Circuit([qf.X(0)])
    quirk_url(circuit_to_quirk(circ), escape=True)


# fin



# Copyright 2016-2018, Rigetti Computing
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

"""
Unit tests for quantumflow.dagcircuit
"""


import numpy as np
import pytest

import quantumflow as qf
from quantumflow.dagcircuit import In, Out


# TODO Refactor in test_circuit
def _test_circ() -> qf.Circuit:
    # Adapted from referenceQVM
    circ = qf.Circuit()
    circ += qf.YPow(1 / 2, 0)
    circ += qf.XPow(1, 0)
    circ += qf.YPow(1 / 2, 1)
    circ += qf.XPow(1, 1)
    circ += qf.CNot(0, 1)
    circ += qf.XPow(-1 / 2, 1)
    circ += qf.YPow(4.71572463191 / np.pi, 1)
    circ += qf.XPow(1 / 2, 1)
    circ += qf.CNot(0, 1)
    circ += qf.XPow(-2 * 2.74973750579 / np.pi, 0)
    circ += qf.XPow(-2 * 2.74973750579 / np.pi, 1)
    return circ


def _true_ket() -> qf.State:
    # Adapted from referenceQVM
    wf_true = np.array(
        [
            0.00167784 + 1.00210180e-05 * 1j,
            0.50000000 - 4.99997185e-01 * 1j,
            0.50000000 - 4.99997185e-01 * 1j,
            0.00167784 + 1.00210180e-05 * 1j,
        ]
    )
    return qf.State(wf_true.reshape((2, 2)))


def test_init() -> None:
    dag = qf.DAGCircuit([])
    assert dag.size() == 0


def test_inverse() -> None:
    dag = qf.DAGCircuit(_test_circ())
    inv_dag = dag.H

    ket0 = qf.random_state(2)
    ket1 = dag.run(ket0)
    ket2 = inv_dag.run(ket1)

    assert qf.states_close(ket0, ket2)


def test_ascircuit() -> None:
    circ0 = qf.ghz_circuit(range(5))
    dag = qf.DAGCircuit(circ0)
    circ1 = qf.Circuit(dag)

    assert tuple(circ1.qubits) == (0, 1, 2, 3, 4)
    assert dag.qubits == circ0.qubits
    assert dag.qubit_nb == 5


def test_str() -> None:
    circ0 = qf.ghz_circuit(range(5))
    dag = qf.DAGCircuit(circ0)
    string = str(dag)
    assert (
        string
        == """DAGCircuit
    H 0
    CNot 0 1
    CNot 1 2
    CNot 2 3
    CNot 3 4\
"""
    )


def test_asgate() -> None:
    circ0 = qf.zyz_circuit(0.1, 2.2, 0.5, 0)
    gate0 = circ0.asgate()
    dag0 = qf.DAGCircuit(circ0)
    gate1 = dag0.asgate()
    assert qf.gates_close(gate0, gate1)


def test_evolve() -> None:
    rho0 = qf.random_state(3).asdensity()
    rho1 = qf.CCNot(0, 1, 2).evolve(rho0)

    dag = qf.DAGCircuit(qf.translate_ccnot_to_cnot(qf.CCNot(0, 1, 2)))
    rho2 = dag.evolve(rho0)

    assert qf.densities_close(rho1, rho2)


def test_aschannel() -> None:
    rho0 = qf.random_state(3).asdensity()
    rho1 = qf.CCNot(0, 1, 2).evolve(rho0)

    dag = qf.DAGCircuit(qf.translate_ccnot_to_cnot(qf.CCNot(0, 1, 2)))
    chan = dag.aschannel()
    rho2 = chan.evolve(rho0)

    assert qf.densities_close(rho1, rho2)


def test_depth() -> None:
    circ = qf.Circuit(qf.QFTGate([0, 1, 2, 3]).decompose())
    dag = qf.DAGCircuit(circ)
    assert dag.depth() == 8

    circ = qf.ghz_circuit(range(5))
    dag = qf.DAGCircuit(circ)
    assert dag.depth() == 5

    assert dag.depth(local=False) == 4


def test_moments() -> None:
    circ0 = qf.ghz_circuit(range(5))
    dag = qf.DAGCircuit(circ0)
    circ = dag.moments()
    assert circ.size() == dag.depth()

    circ1 = qf.Circuit(
        [
            qf.Z(0),
            qf.Z(1),
            qf.Z(2),
            qf.CNot(0, 1),
            qf.Measure(0, 0),
            qf.Measure(1, 1),
            qf.Measure(2, 2),
        ]
    )
    moments = qf.DAGCircuit(circ1).moments()
    print()
    print(moments)

    assert len(moments) == 3
    assert len(moments[0]) == 3  # type: ignore
    assert len(moments[1]) == 1  # type: ignore
    assert len(moments[2]) == 3  # type: ignore

    with pytest.warns(DeprecationWarning):
        circ = dag.layers()


def test_components() -> None:
    circ = qf.Circuit()
    circ += qf.H(0)
    circ += qf.H(1)
    dag = qf.DAGCircuit(circ)
    assert dag.component_nb() == 2

    circ += qf.CNot(0, 1)
    dag = qf.DAGCircuit(circ)
    assert dag.component_nb() == 1

    circ0 = qf.ghz_circuit([0, 2, 4, 6, 8])
    circ1 = qf.ghz_circuit([1, 3, 5, 7, 9])

    circ = circ0 + circ1
    dag = qf.DAGCircuit(circ)
    comps = dag.components()
    assert dag.component_nb() == 2

    circ0 = qf.Circuit(qf.QFTGate([0, 2, 4, 6]).decompose())
    circ1 = qf.ghz_circuit([1, 3, 5, 7])
    circ += circ0
    circ += circ1
    circ += qf.H(10)
    dag = qf.DAGCircuit(circ)
    comps = dag.components()
    assert dag.component_nb() == 3
    assert len(comps) == 3


def test_next_prev() -> None:
    circ = qf.ghz_circuit([0, 2, 4, 6, 8])
    elem = circ[3]
    dag = qf.DAGCircuit(circ)

    assert dag.next_element(elem, elem.qubits[1]) == circ[4]
    assert dag.prev_element(elem, elem.qubits[0]) == circ[2]

    assert dag.next_element(elem, elem.qubits[0]) == Out(4)
    assert dag.prev_element(elem, elem.qubits[1]) == In(6)


def test_on() -> None:
    circ = qf.Circuit()
    circ += qf.H(0)
    circ += qf.H(1)
    dag = qf.DAGCircuit(circ)

    dag = dag.on(2, 3)
    assert dag.qubits == (2, 3)

    dag = dag.rewire({2: 4, 3: 6})
    assert dag.qubits == (4, 6)

    with pytest.raises(ValueError):
        dag.on(2, 3, 5)


# fin



# Copyright 2019-, Gavin E. Crooks and the QuantumFlow contributors
# Copyright 2016-2018, Rigetti Computing
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

"""
.. contents:: :local:
.. currentmodule:: quantumflow

QuantumFlow module for working with the Pauli algebra.

.. autoclass:: Pauli
    :members:

.. autofunction:: sX
.. autofunction:: sY
.. autofunction:: sZ
.. autofunction:: sI

.. autofunction:: pauli_sum
.. autofunction:: pauli_product
.. autofunction:: pauli_pow
.. autofunction:: paulis_commute
.. autofunction:: pauli_commuting_sets
.. autofunction:: paulis_close

"""

# Kudos: Adapted from PyQuil's paulis.py, original written by Nick Rubin

import heapq
from cmath import isclose  # type: ignore
from functools import reduce
from itertools import groupby, product
from numbers import Complex
from operator import itemgetter, mul
from typing import Any, Dict, Iterator, List, Set, Tuple

import numpy as np

from . import var
from .config import ATOL
from .ops import Operation
from .qubits import Qubit, Qubits
from .states import State
from .tensors import TensorLike
from .var import Variable

__all__ = [
    "PauliTerm",
    "Pauli",
    "sX",
    "sY",
    "sZ",
    "sI",
    "pauli_sum",
    "pauli_product",
    "pauli_pow",
    "paulis_commute",
    "pauli_commuting_sets",
    "paulis_close",
    "pauli_decompose_hermitian",
]


PauliTerm = Tuple[Qubits, str, Variable]

PAULI_OPS = ["X", "Y", "Z", "I"]

PAULI_PROD = {
    "ZZ": ("I", 1.0),
    "YY": ("I", 1.0),
    "XX": ("I", 1.0),
    "II": ("I", 1.0),
    "XY": ("Z", 1.0j),
    "XZ": ("Y", -1.0j),
    "YX": ("Z", -1.0j),
    "YZ": ("X", 1.0j),
    "ZX": ("Y", 1.0j),
    "ZY": ("X", -1.0j),
    "IX": ("X", 1.0),
    "IY": ("Y", 1.0),
    "IZ": ("Z", 1.0),
    "ZI": ("Z", 1.0),
    "YI": ("Y", 1.0),
    "XI": ("X", 1.0),
}


class Pauli(Operation):
    """
    An element of the Pauli algebra.

    An element of the Pauli algebra is a sequence of terms, such as

        Y(1) - 0.5 Z(1) X(2) Y(4)

    where X, Y, Z and I are the 1-qubit Pauli operators.

    """

    # Internally, each term is a tuple of a complex coefficient, and a sequence
    # of single qubit Pauli operators. (The coefficient goes last so that the
    # terms sort on the operators).
    #
    # PauliTerm = Tuple[Tuple[Tuple[Qubit, str], ...], complex]
    #
    # Each Pauli operator consists of a tuple of
    # qubits e.g. (0, 1, 3), a tuple of Pauli operators e.g. ('X', 'Y', 'Z').
    # Qubits and Pauli terms are kept in sorted order. This ensures that a
    # Pauli element has a unique representation, and makes summation and
    # simplification efficient. We use Tuples (and not lists) because they are
    # immutable and hashable.

    def __init__(self, *terms: PauliTerm) -> None:
        the_terms = []
        qubits: Set[Qubit] = set()

        if len(terms) != 0:  # == 0 zero element
            # print(terms)
            for qbs, ops, coeff in terms:
                if not all(op in PAULI_OPS for op in ops):
                    raise ValueError("Valid Pauli operators are I, X, Y, and Z")
                if isinstance(coeff, Complex) and isclose(coeff, 0.0):
                    continue
                if len(ops) != 0:
                    qops = sorted(zip(qbs, ops))
                    qops = list(filter(lambda x: x[1] != "I", qops))
                    qbs, ops = zip(*qops) if qops else ((), "")  # type: ignore

                the_terms.append((tuple(qbs), "".join(ops), coeff))
                qubits.update(qbs)

        qbs = sorted(list(qubits))
        super().__init__(qbs)

        self.terms = tuple(the_terms)

    # Rename coeff?
    @classmethod
    def term(cls, qubits: Qubits, ops: str, coefficient: Variable = 1.0) -> "Pauli":
        """
        Create an element of the Pauli algebra from a sequence of qubits
        and operators. Qubits must be unique and sortable
        """
        return cls((qubits, ops, coefficient))

    @classmethod
    def sigma(cls, qubit: Qubit, operator: str, coefficient: complex = 1.0) -> "Pauli":
        """Returns a Pauli operator ('I', 'X', 'Y', or 'Z') acting
        on the given qubit"""
        if operator == "I":
            return cls.scalar(coefficient)
        return cls.term([qubit], operator, coefficient)

    @classmethod
    def scalar(cls, coefficient: complex) -> "Pauli":
        """Return a scalar multiple of the Pauli identity element."""
        return cls.term((), "", coefficient)

    def is_scalar(self) -> bool:
        """Returns true if this object is a scalar multiple of the Pauli
        identity element"""
        if len(self.terms) > 1:
            return False
        if len(self.terms) == 0:
            return True  # Zero element
        if len(self.terms[0][0]) == 0:
            return True
        return False

    @classmethod
    def identity(cls) -> "Pauli":
        """Return the identity element of the Pauli algebra"""
        return cls.scalar(1.0)

    def is_identity(self) -> bool:
        """Returns True if this object is identity Pauli element."""

        if len(self) != 1:
            return False
        if self.terms[0][0] != ():
            return False
        return isclose(self.terms[0][2], 1.0)  # FIXME Variables

    @classmethod
    def zero(cls) -> "Pauli":
        """Return the zero element of the Pauli algebra"""
        return cls()

    def is_zero(self) -> bool:
        """Return True if this object is the zero Pauli element."""
        return len(self.terms) == 0

    # FIXME?
    # @property
    # def qubits(self) -> Qubits:
    #     """Return a list of qubits acted upon by the Pauli element"""
    #     return list({q for term, _ in self.terms
    #                  for q, _ in term})  # type: ignore

    def __repr__(self) -> str:
        return "Pauli(" + str(self.terms) + ")"

    def __str__(self) -> str:
        out = []
        for qbs, ops, coeff in self.terms:
            str_coeff = str(coeff)
            if str_coeff[0] != "-":
                str_coeff = "+" + str_coeff

            if len(ops) == 0:
                # scalar
                out.append(str_coeff)
            else:
                if coeff == 1:
                    out.append("+")
                elif coeff == -1:
                    out.append("-")
                else:
                    out.append(str_coeff)

                out.append(" * ".join(f"s{op}({q})" for q, op in zip(qbs, ops)))

        return " ".join(out)

    # FIXME
    # TODO: Should return Paulis if at all
    def __iter__(self) -> Iterator[PauliTerm]:  # type: ignore  # FIXME
        return iter(self.terms)

    def __len__(self) -> int:
        return len(self.terms)

    def __add__(self, other: Any) -> "Pauli":
        if isinstance(other, Complex):
            other = Pauli.scalar(complex(other))
        return pauli_sum(self, other)

    def __radd__(self, other: Any) -> "Pauli":
        return self.__add__(other)

    def __mul__(self, other: Any) -> "Pauli":
        if var.is_symbolic(other):
            other = Pauli.scalar(other)
        if isinstance(other, Complex):
            other = Pauli.scalar(complex(other))
        return pauli_product(self, other)

    def __rmul__(self, other: Any) -> "Pauli":
        return self.__mul__(other)

    def __truediv__(self, other: Any) -> "Pauli":
        return self * (1 / other)

    def __sub__(self, other: Any) -> "Pauli":
        return self + -1 * other

    def __rsub__(self, other: Any) -> "Pauli":
        return other + -1 * self

    def __neg__(self) -> "Pauli":
        return self * -1

    def __pos__(self) -> "Pauli":
        return self

    def __pow__(self, exponent: int) -> "Pauli":
        return pauli_pow(self, exponent)

    # Note: functools.total_ordering does not play nice with mypy
    # So we add all comparisons explicitly

    def __lt__(self, other: Any) -> bool:
        if not isinstance(other, Pauli):
            return NotImplemented
        return self.terms < other.terms

    def __le__(self, other: Any) -> bool:
        if not isinstance(other, Pauli):
            return NotImplemented
        return self.terms <= other.terms

    def __gt__(self, other: Any) -> bool:
        if not isinstance(other, Pauli):
            return NotImplemented
        return self.terms > other.terms

    def __ge__(self, other: Any) -> bool:
        if not isinstance(other, Pauli):
            return NotImplemented
        return self.terms >= other.terms

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Pauli):
            return NotImplemented
        return self.terms == other.terms

    def __hash__(self) -> int:
        return hash(self.terms)

    def asoperator(self, qubits: Qubits = None) -> TensorLike:
        # DOCME: Use of qubits argument here.

        # Late import to prevent circular imports
        from .modules import IdentityGate
        from .ops import StdGate

        NAMED_GATES = StdGate.cv_stdgates

        qubits = self.qubits if qubits is None else qubits
        if self.is_zero():
            N = len(qubits)
            return np.zeros(shape=(2 ** N, 2 ** N))

        res = []
        for qbs, ops, coeff in self.terms:
            if var.is_symbolic(coeff):
                coeff = complex(coeff)
            gate = IdentityGate(qubits)
            for q, op in zip(qbs, ops):
                gate = NAMED_GATES[op](q) @ gate  # type: ignore
            res.append(gate.asoperator() * coeff)
        return sum(res)

    # TESTME
    def run(self, ket: State) -> State:
        from .ops import StdGate

        NAMED_GATES = StdGate.cv_stdgates

        resultants = []
        for qbs, ops, coeff in self.terms:
            res = State(ket.tensor * coeff, ket.qubits)
            for q, op in zip(qbs, ops):
                res = NAMED_GATES[op](q).run(res)  # type: ignore
            resultants.append(res.tensor)

        out = State(sum(resultants), ket.qubits)
        return out

    # TESTME (was broken)
    def on(self, *qubits: Qubit) -> "Pauli":
        if len(qubits) != self.qubit_nb:
            raise ValueError("Wrong number of qubits")

        return self.rewire(dict(zip(self.qubits, qubits)))

    def rewire(self, labels: Dict[Qubit, Qubit]) -> "Pauli":
        paulis = []
        for qbs, ops, coeff in self.terms:
            new_qbs = [labels[q] for q in qbs]
            paulis.append(Pauli.term(new_qbs, ops, coeff))

        return pauli_sum(*paulis)


# End class Pauli


def sX(qubit: Qubit, coefficient: complex = 1) -> Pauli:
    """Return the Pauli sigma_X operator acting on the given qubit"""
    return Pauli.sigma(qubit, "X", coefficient)


def sY(qubit: Qubit, coefficient: complex = 1) -> Pauli:
    """Return the Pauli sigma_Y operator acting on the given qubit"""
    return Pauli.sigma(qubit, "Y", coefficient)


def sZ(qubit: Qubit, coefficient: complex = 1) -> Pauli:
    """Return the Pauli sigma_Z operator acting on the given qubit"""
    return Pauli.sigma(qubit, "Z", coefficient)


def sI(qubit: Qubit, coefficient: complex = 1) -> Pauli:
    """Return the Pauli sigma_I (identity) operator. The qubit is irrelevant,
    but kept as an argument for consistency"""
    return Pauli.sigma(qubit, "I", coefficient)


def pauli_sum(*elements: Pauli) -> Pauli:
    """Return the sum of elements of the Pauli algebra"""
    terms = []

    key = itemgetter(0, 1)  # (qbs, str)
    for term, grp in groupby(heapq.merge(*elements, key=key), key=key):
        coeff = sum(g[2] for g in grp)
        if not var.almost_zero(coeff):
            terms.append((term[0], term[1], coeff))

    return Pauli(*terms)


def pauli_product(*elements: Pauli) -> Pauli:
    """Return the product of elements of the Pauli algebra"""
    result_terms = []

    for terms in product(*elements):
        coeff = reduce(mul, [term[2] for term in terms])
        ops = (zip(qbs, ops) for qbs, ops, _ in terms)
        out_qubits = []
        out_ops = []
        key = itemgetter(0)
        for qubit, qops in groupby(heapq.merge(*ops, key=key), key=key):
            res = next(qops)[1]  # Operator: X Y Z
            for op in qops:
                pair = res + op[1]
                res, rescoeff = PAULI_PROD[pair]
                coeff *= rescoeff
            if res != "I":
                out_qubits.append(qubit)
                out_ops.append(res)

        p = Pauli.term(out_qubits, "".join(out_ops), coeff)
        result_terms.append(p)

    return pauli_sum(*result_terms)


# # FIXME DOCME TESTME
# def pauli_grad(pauli: Pauli, *x: Variable) -> Pauli:
#     result_terms = []

#     for term, coeff in pauli:
#         d = sympy.diff(coeff, *x)
#         p = Pauli(((term, d),))
#         if not d.is_zero:
#             # print(d, p)
#             result_terms.append(p)
#     return pauli_sum(*result_terms)


def pauli_pow(pauli: Pauli, exponent: int) -> Pauli:
    """
    Raise an element of the Pauli algebra to a non-negative integer power.
    """

    if not isinstance(exponent, int) or exponent < 0:
        raise ValueError("The exponent must be a non-negative integer.")

    if exponent == 0:
        return Pauli.identity()

    if exponent == 1:
        return pauli

    # https://en.wikipedia.org/wiki/Exponentiation_by_squaring
    y = Pauli.identity()
    x = pauli
    n = exponent
    while n > 1:
        if n % 2 == 0:  # Even
            x = x * x
            n = n // 2
        else:  # Odd
            y = x * y
            x = x * x
            n = (n - 1) // 2
    return x * y


def paulis_close(pauli0: Pauli, pauli1: Pauli, atol: float = ATOL) -> bool:
    """Returns: True if Pauli elements are almost identical."""
    pauli = pauli0 - pauli1
    d = sum(abs(coeff) ** 2 for _, _, coeff in pauli.terms)
    return d <= atol


def paulis_commute(element0: Pauli, element1: Pauli) -> bool:
    """
    Return true if the two elements of the Pauli algebra commute.
    i.e. if element0 * element1 == element1 * element0

    Derivation similar to arXiv:1405.5749v2 for the check_commutation step in
    the Raeisi, Wiebe, Sanders algorithm (arXiv:1108.4318, 2011).
    """

    def _coincident_parity(term0: PauliTerm, term1: PauliTerm) -> bool:
        non_similar = 0
        key = itemgetter(0)

        op0 = zip(term0[0], term0[1])
        op1 = zip(term1[0], term1[1])
        for _, qops in groupby(heapq.merge(op0, op1, key=key), key=key):

            listqops = list(qops)
            if len(listqops) == 2 and listqops[0][1] != listqops[1][1]:
                non_similar += 1
        return non_similar % 2 == 0

    for term0, term1 in product(element0, element1):
        if not _coincident_parity(term0, term1):
            return False

    return True


def pauli_commuting_sets(element: Pauli) -> Tuple[Pauli, ...]:
    """Gather the terms of a Pauli polynomial into commuting sets.

    Uses the algorithm defined in (Raeisi, Wiebe, Sanders,
    arXiv:1108.4318, 2011) to find commuting sets. Except uses commutation
    check from arXiv:1405.5749v2
    """
    if len(element) < 2:
        return (element,)

    groups: List[Pauli] = []

    for term in element:
        pterm = Pauli.term(*term)

        assigned = False
        for i, grp in enumerate(groups):
            if paulis_commute(grp, pterm):
                groups[i] = grp + pterm
                assigned = True
                break
        if not assigned:
            groups.append(pterm)

    return tuple(groups)


def pauli_decompose_hermitian(matrix: TensorLike, qubits: Qubits = None) -> Pauli:
    """Decompose a Hermitian matrix into an element of the Pauli algebra.

    This works because tensor products of Pauli matrices form an orthonormal
    basis in the linear space of all 2^N×2^N matrices under Hilbert-Schmidt
    inner product.
    """
    # TODO: This should work for any matrix, not just Hermitian?
    #       Generalize: remove hermitian check and coercing coefficients to real.

    if not np.ndim(matrix) == 2:
        raise ValueError("Must be square matrix")

    # Wait is this true?
    if not np.allclose(matrix.conj().T, matrix):
        raise ValueError("Matrix must be Hermitian")

    N = int(np.log2(np.size(matrix))) // 2
    if not 2 ** (2 * N) == np.size(matrix):
        raise ValueError("Matrix dimensions must be power of 2")

    if qubits is None:
        qubits = list(range(N))
    else:
        assert len(qubits) == N

    terms = []
    for ops in product("IXYZ", repeat=N):
        P = Pauli.term(qubits, "".join(ops)).asoperator(qubits=qubits)
        coeff = np.real(np.trace(P @ matrix) / (2 ** N))
        term = Pauli.term(qubits, "".join(ops), coeff)
        terms.append(term)

    return pauli_sum(*terms)


# fin



# Copyright 2016-2018, Rigetti Computing
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

"""
QuantumFlow: Directed Acyclic Graph representations of a Circuit.
"""

import itertools
import textwrap
from typing import Any, Dict, Generator, Iterable, Iterator, List, Tuple

import networkx as nx
import numpy as np
import opt_einsum

from . import utils
from .circuits import Circuit
from .config import CIRCUIT_INDENT
from .ops import Channel, Gate, Operation, Unitary
from .qubits import Qubit, Qubits
from .states import Density, State
from .stdops import Moment

__all__ = ("DAGCircuit",)


# DOCME
class In(Operation):
    def __init__(self, q0: Qubit):
        super().__init__([q0])

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, In) and other.qubits == self.qubits

    def __hash__(self) -> int:
        return hash(self.qubits)


# DOCME
class Out(Operation):
    def __init__(self, q0: Qubit):
        super().__init__([q0])

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, Out) and other.qubits == self.qubits

    def __hash__(self) -> int:
        return hash(self.qubits)


# FIXME: Design flaw!?
# DAGCircuit fails if we try to add multi instances of the same gate
# The problem is that we use gates as nodes in the graph, so they have to be
# unique.
# Either wrap all Operations in some QFNode class before adding to circuit,
# Or make sure all Operations occur only once in circuit.
# Clone copies as necessary.


class DAGCircuit(Operation):
    """A Directed Acyclic Graph representation of a Circuit.

    The circuit is converted to a networkx directed acyclic multi-graph,
    stored in the `graph` attribute.

    There are 3 node types, 'in' nodes representing qubits at the start of a
    circuit; operation nodes; and 'out' nodes for qubits at the
    end of a circuit. Edges are directed from 'in' to 'out' via the Operation
    nodes. Each edge is keyed to a qubit.

    A DAGCircuit is considered a mutable object. But it should not be mutated
    if it is part of another Circuit or other immutable composite Operation.

    DAGCircuit is iterable, yielding all of the operation nodes in
    topological sort order.

    Note: Provisional API
    """

    def __init__(self, elements: Iterable[Operation]) -> None:

        self.graph = nx.MultiDiGraph()
        self._qubits_in: Dict[Qubit, In] = {}
        self._qubits_out: Dict[Qubit, Out] = {}

        for elem in elements:
            self.append(elem)  # type: ignore

    def append(self, elem: Operation) -> None:
        G = self.graph
        G.add_node(elem)
        for qubit in elem.qubits:
            if qubit not in self._qubits_in:
                qin = In(qubit)
                qout = Out(qubit)
                self._qubits_in[qubit] = qin
                self._qubits_out[qubit] = qout
                G.add_edge(qin, qout, key=qubit)
            qout = self._qubits_out[qubit]
            prev = list(G.predecessors(qout))[0]
            G.remove_edge(prev, qout)
            G.add_edge(prev, elem, key=qubit)
            G.add_edge(elem, qout, key=qubit)

    @property
    def qubits(self) -> Qubits:
        qbs = list(self._qubits_in.keys())
        qbs = sorted(qbs)
        return tuple(qbs)

    @property
    def qubit_nb(self) -> int:
        return len(self.qubits)

    def on(self, *qubits: Qubit) -> "DAGCircuit":
        return DAGCircuit(Circuit(self).on(*qubits))

    def rewire(self, labels: Dict[Qubit, Qubit]) -> "DAGCircuit":
        return DAGCircuit(Circuit(self).rewire(labels))

    @property
    def H(self) -> "DAGCircuit":
        return DAGCircuit(Circuit(self).H)

    def run(self, ket: State) -> State:
        for elem in self:
            ket = elem.run(ket)
        return ket

    def evolve(self, rho: Density) -> Density:
        for elem in self:
            rho = elem.evolve(rho)
        return rho

    def asgate(self) -> Gate:
        # Note: Experimental
        # Contract entire tensor network graph using einsum
        # TODO: Do same for run() and evolve() and aschannel()
        # Note: Does not work with CTF backend
        tensors = []
        sublists = []

        for elem in self:
            assert isinstance(elem, Gate)
            tensors.append(elem.tensor)
            sublists.append(self.next_edges(elem) + self.prev_edges(elem))

        tensors_sublists = list(itertools.chain(*zip(tensors, sublists)))

        for qout in self._qubits_out.values():
            outsubs = self.prev_edges(qout)
        for qin in self._qubits_in.values():
            outsubs.extend(self.next_edges(qin))
        tensors_sublists.append(tuple(outsubs))

        tensor = opt_einsum.contract(*tensors_sublists)

        return Unitary(tensor, self.qubits)

    def aschannel(self) -> Channel:
        return Circuit(self).aschannel()

    def depth(self, local: bool = True) -> int:
        """Return the circuit depth.

        Args:
            local:  If True include local one-qubit gates in depth
                calculation. Else return the multi-qubit gate depth.
        """
        G = self.graph
        if not local:

            def remove_local(dagc: DAGCircuit) -> Generator[Operation, None, None]:
                for elem in dagc:
                    if dagc.graph.degree[elem] > 2:
                        yield elem

            G = DAGCircuit(remove_local(self)).graph

        return nx.dag_longest_path_length(G) - 1

    def size(self) -> int:
        """Return the number of operations."""
        return self.graph.order() - 2 * self.qubit_nb

    def component_nb(self) -> int:
        """Return the number of independent components that this
        DAGCircuit can be split into."""
        return nx.number_weakly_connected_components(self.graph)

    def components(self) -> List["DAGCircuit"]:
        """Split DAGCircuit into independent components"""
        G = self.graph
        comps = (G.subgraph(c).copy() for c in nx.weakly_connected_components(G))
        return [DAGCircuit(comp) for comp in comps]

    def moments(self) -> Circuit:
        """Split DAGCircuit into Moments, where the operations within each
        moment operate on different qubits (and therefore commute).

        Returns: A Circuit of Moments
        """
        D = self.depth()
        G = self.graph

        node_depth: Dict[Qubit, int] = {}
        node_height: Dict[Qubit, int] = {}

        for elem in self:
            depth = np.max(
                list(node_depth.get(prev, -1) + 1 for prev in G.predecessors(elem))
            )
            node_depth[elem] = depth

        for elem in reversed(list(self)):
            height = np.min(
                list(node_height.get(next, D) - 1 for next in G.successors(elem))
            )
            node_height[elem] = height

        # Place operations in Moment closest to
        # beginning or end of circuit.
        moments = [Circuit() for _ in range(D)]

        # Iterate nodes in reverse seems to preserve original circuit ordering
        for elem in reversed(list(self)):
            depth = node_depth[elem]
            height = node_height[elem]
            if depth <= D - height - 1:
                moments[depth] += elem
            else:
                moments[height] += elem

        circ = Circuit(Moment(moment) for moment in moments)

        return circ

    @utils.deprecated
    def layers(self) -> Circuit:
        return self.moments()

    def __iter__(self) -> Iterator[Operation]:
        for elem in nx.topological_sort(self.graph):
            # Filter in and out nodes
            if not isinstance(elem, In) and not isinstance(elem, Out):
                yield elem

    # DOCME TESTME
    def next_element(self, elem: Operation, qubit: Qubit = None) -> Operation:
        for _, node, key in self.graph.edges(elem, keys=True):
            if qubit is None or key == qubit:
                return node
        assert False  # Insanity check  # FIXME, raise exception

    # DOCME TESTME
    def prev_element(self, elem: Operation, qubit: Qubit = None) -> Operation:
        for node, _, key in self.graph.in_edges(elem, keys=True):
            if qubit is None or key == qubit:
                return node
        assert False  # Insanity check  # FIXME, raise exception

    def next_edges(self, elem: Operation) -> List[Tuple[Any, Any, Qubit]]:
        qubits = elem.qubits
        N = len(qubits)
        edges = [None] * N
        for edge in self.graph.out_edges(elem, keys=True):
            edges[qubits.index(edge[2])] = edge

        return list(edges)  # type: ignore

    def prev_edges(self, elem: Operation) -> List[Tuple[Any, Any, Qubit]]:
        qubits = elem.qubits
        N = len(qubits)
        edges = [None] * N
        for edge in self.graph.in_edges(elem, keys=True):
            edges[qubits.index(edge[2])] = edge

        return list(edges)  # type: ignore

    def __str__(self) -> str:
        circ_str = "\n".join([str(elem) for elem in self])
        circ_str = textwrap.indent(circ_str, " " * CIRCUIT_INDENT)
        return "\n".join([self.name, circ_str])


# End class DAGCircuit



# Copyright 2020-, Gavin E. Crooks and contributors
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

# DOCME

import string
from typing import Any, List, Sequence, Tuple

import numpy as np

__all__ = ("QubitTensor", "TensorLike", "asqutensor")


qubit_dtype = np.complex128
"""The complex data type used by the backend"""

QubitTensor = Any
"""Type hint for numpy arrays representing quantum data.
QubitTensor arrays have complex data type, and all axes have length 2."""

# NB: Currently only used for documentation, not type checking


TensorLike = Any
"""Any python object that can be converted into a numpy array"""

# NB: This annotation currently only used for documentation, not type checking


def asqutensor(array: TensorLike) -> QubitTensor:
    """Converts a tensor like object to a numpy array object with complex data
    type, and reshapes to [2]*N (So the number of elements must be a power of 2)
    """
    tensor = np.asarray(array, dtype=qubit_dtype)

    N = int(np.log2(np.size(tensor)))
    shape = (2,) * N

    if tensor.shape != shape:  # Only reshape if necessary
        tensor = tensor.reshape(shape)

    return tensor


EINSUM_SUBSCRIPTS = string.ascii_lowercase + string.ascii_uppercase
"""Subscripts that can be used with numpy's einsum"""


def flatten(tensor: QubitTensor, rank: int) -> np.ndarray:
    """Return tensor with with qubit indices flattened"""
    R = rank
    N = np.ndim(tensor) // R
    return np.reshape(tensor, [2 ** N] * R)


def transpose(tensor: QubitTensor, perm: Sequence[int] = None) -> QubitTensor:
    """(Super)-operator transpose. Permutes the meta-indices.
    Default is to invert the meta-index order.
    """
    R = len(perm) if perm is not None else 2
    tensor = flatten(tensor, R)
    tensor = np.transpose(tensor, perm)
    tensor = asqutensor(tensor)

    return tensor


def conj_transpose(tensor: QubitTensor) -> QubitTensor:
    """Return the conjugate transpose of this tensor."""
    tensor = transpose(tensor)
    tensor = np.conj(tensor)
    return tensor


def permute(tensor: QubitTensor, perm: Sequence[int]) -> QubitTensor:
    """Permute the order of the subindexes"""
    N = len(perm)
    R = np.ndim(tensor) // N

    pperm: List[int] = []
    for rr in range(0, R):
        pperm += [rr * N + idx for idx in perm]
    tensor = np.transpose(tensor, pperm)

    return tensor


def inner(tensor0: QubitTensor, tensor1: QubitTensor) -> QubitTensor:
    """Return the inner product between two meta tensors
    (Assuming same meta-shape)"""
    # Note: Relies on fact that vdot flattens arrays
    return np.vdot(tensor0, tensor1)  # type: ignore


def outer(tensor0: QubitTensor, tensor1: QubitTensor, rank: int) -> QubitTensor:
    """Outer (direct) product of qubit tensors"""
    R = rank
    N0 = np.ndim(tensor0) // R
    N1 = np.ndim(tensor1) // R

    tensor = np.outer(tensor0, tensor1)

    # Interleave meta axes
    # R = 1  perm = (0, 1)
    # R = 2  perm = (0, 2, 1, 3)
    # R = 4  perm = (0, 4, 1, 5, 2, 6, 3, 7)
    tensor = np.reshape(tensor, ([2 ** N0] * R) + ([2 ** N1] * R))
    perm = [idx for ij in zip(range(0, R), range(R, 2 * R)) for idx in ij]
    if R != 1:
        tensor = transpose(tensor, perm)

    return tensor


def trace(tensor: QubitTensor, rank: int) -> float:
    """
    Return the trace, the sum of the diagonal elements of the (super)
    operator.
    """
    R = rank
    N = np.ndim(tensor) // R

    if R == 1:
        raise ValueError("Cannot take trace of vector")  # pragma: no cover

    tensor = np.reshape(tensor, [2 ** (N * R // 2)] * 2)
    trace = np.trace(tensor)
    return trace


def norm(tensor: QubitTensor) -> QubitTensor:
    """Return the norm of this vector"""
    return np.absolute(inner(tensor, tensor))


def diag(tensor: QubitTensor) -> QubitTensor:
    """Returns the matrix diagonal of the product tensor"""
    tensor = flatten(tensor, 2)
    tensor = np.diag(tensor)
    tensor = asqutensor(tensor)

    return tensor


# TESTME
# TESTME on channels and density
def partial_trace(
    tensor: QubitTensor, indices: Sequence[int], rank: int = 2
) -> QubitTensor:
    """
    Return the partial trace over some subset of qubits,

    Args:
        tensor: A QubitTensor
        indices: The set of qubits that should be kept.
        rank: Rank of qubit tensor.
    """
    R = rank
    N = np.ndim(tensor) // R

    if R == 1:
        raise ValueError("Cannot take trace of vector")  # pragma: no cover

    subscripts = list(EINSUM_SUBSCRIPTS[0 : N * R])

    print(indices, R, N, subscripts)
    for idx in indices:
        for r in range(1, R):
            subscripts[r * N + idx] = subscripts[idx]
    subscript_str = "".join(subscripts)

    # Only numpy's einsum works with repeated subscripts
    tensor = np.einsum(subscript_str, tensor)

    return tensor


# DOCME
def tensormul(
    tensor0: QubitTensor,
    tensor1: QubitTensor,
    indices: Tuple[int, ...],
    diagonal: bool = False,
) -> QubitTensor:
    N = np.ndim(tensor1)
    K = np.ndim(tensor0) // 2
    assert K == len(indices)

    gate = np.reshape(tensor0, [2 ** K, 2 ** K])

    perm = list(indices) + [n for n in range(N) if n not in indices]
    inv_perm = np.argsort(perm)

    tensor = tensor1
    tensor = np.transpose(tensor, perm)
    tensor = np.reshape(tensor, [2 ** K, 2 ** (N - K)])

    if diagonal:
        tensor = np.transpose(tensor)
        tensor = tensor * np.diag(gate)
        tensor = np.transpose(tensor)
    else:
        tensor = np.matmul(gate, tensor)

    tensor = np.reshape(tensor, [2] * N)
    tensor = np.transpose(tensor, inv_perm)

    return tensor


# fin



# Copyright 2020-, Gavin E. Crooks and contributors
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

"""
.. module:: quantumflow
.. contents:: :local:
.. currentmodule:: quantumflow

Qubits
======

States, gates, and various other methods accept a list of qubits labels upon
which the given State or Gate acts. A Qubit label can be any hashable python
object, but typically an integer or string. e.g. `[0, 1, 2]`, or
`['a', 'b', 'c']`. Note that some operations expect the qubits to be sortable,
so don't mix different incomparable data types.

"""


from typing import Any, Sequence

__all__ = ("Qubit", "Qubits")


Qubit = Any
"""
Type for qubits. Any hashable python object.
Qubits should be mutually sortable, so don't mix different types, e.g. integers
and strings.
"""
# This used to be 'Qubit = Hashable', but mypy started complaining that
# you cant sort Hashable objects. There doesn't seem to be any good way of
# specifying a type that's sortable and hashable.


Qubits = Sequence[Qubit]
"""Type for sequence of qubits"""


# fin



# Copyright 2019-, Gavin E. Crooks and contributors
# Copyright 2016-2018, Rigetti Computing
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

"""
QuantumFlow Gate Decompositions

.. contents:: :local:
.. currentmodule:: quantumflow


One-qubit gate decompositions
#############################

.. autofunction:: bloch_decomposition
.. autofunction:: zyz_decomposition
.. autofunction:: euler_decomposition

Two-qubit gate decompositions
#############################

.. autofunction:: kronecker_decomposition
.. autofunction:: canonical_decomposition
.. autofunction:: canonical_coords
.. autofunction:: cnot_decomposition
.. autofunction:: b_decomposition
"""


import itertools
from typing import List, Sequence, Tuple, cast

import numpy as np

from .circuits import Circuit, euler_circuit
from .config import ATOL
from .info import gates_close
from .ops import Gate, Unitary
from .stdgates import S_H, B, Can, I, Rn, S, V, X, Y, YPow, Z, ZPow
from .translate import translate_can_to_cnot

__all__ = [
    "bloch_decomposition",
    "zyz_decomposition",
    "euler_decomposition",
    "kronecker_decomposition",
    "canonical_decomposition",
    "canonical_coords",
    "cnot_decomposition",
    "b_decomposition",
    "convert_can_to_weyl",
]


# TODO: Optionally include phase
def bloch_decomposition(gate: Gate) -> Circuit:
    """
    Converts a 1-qubit gate into a RN gate, a 1-qubit rotation of angle theta
    about axis (nx, ny, nz) in the Bloch sphere.

    Returns:
        A Circuit containing a single RN gate
    """
    if gate.qubit_nb != 1:
        raise ValueError("Expected 1-qubit gate")

    U = gate.asoperator()
    U /= np.linalg.det(U) ** (1 / 2)

    nx = -U[0, 1].imag
    ny = -U[0, 1].real
    nz = -U[0, 0].imag
    N = np.sqrt(nx ** 2 + ny ** 2 + nz ** 2)
    if N == 0:  # Identity
        nx, ny, nz = 1, 1, 1
    else:
        nx /= N
        ny /= N
        nz /= N
    sin_halftheta = N
    cos_halftheta = U[0, 0].real
    theta = 2 * np.arctan2(sin_halftheta, cos_halftheta)

    # We return a Circuit (rather than just a gate) to keep the
    # interface of decomposition routines uniform.
    return Circuit([Rn(theta, nx, ny, nz, *gate.qubits)])


# TODO: Optionally include phase?
def zyz_decomposition(gate: Gate) -> Circuit:
    """
    Returns the Euler Z-Y-Z decomposition of a local 1-qubit gate.
    """
    if gate.qubit_nb != 1:
        raise ValueError("Expected 1-qubit gate")

    (q,) = gate.qubits

    U = gate.su().asoperator()  # SU(2)

    if abs(U[0, 0]) > abs(U[1, 0]):
        theta1 = 2 * np.arccos(min(abs(U[0, 0]), 1))
    else:
        theta1 = 2 * np.arcsin(min(abs(U[1, 0]), 1))

    cos_halftheta1 = np.cos(theta1 / 2)
    if not np.isclose(cos_halftheta1, 0.0):
        phase = U[1, 1] / cos_halftheta1
        theta0_plus_theta2 = 2 * np.arctan2(np.imag(phase), np.real(phase))
    else:
        theta0_plus_theta2 = 0.0

    sin_halftheta1 = np.sin(theta1 / 2)
    if not np.isclose(sin_halftheta1, 0.0):
        phase = U[1, 0] / sin_halftheta1
        theta0_sub_theta2 = 2 * np.arctan2(np.imag(phase), np.real(phase))
    else:
        theta0_sub_theta2 = 0.0

    theta0 = (theta0_plus_theta2 + theta0_sub_theta2) / 2
    theta2 = (theta0_plus_theta2 - theta0_sub_theta2) / 2

    t0 = theta0 / np.pi
    t1 = theta1 / np.pi
    t2 = theta2 / np.pi

    if np.isclose(t1, 0.0):
        t2 = t0 + t2
        t1 = 0.0
        t0 = 0.0

    circ1 = Circuit()
    circ1 += ZPow(t2, q)
    circ1 += YPow(t1, q)
    circ1 += ZPow(t0, q)

    return circ1


def euler_decomposition(gate: Gate, euler: str = "ZYZ") -> Circuit:
    """
    Returns an Euler angle decomposition of a local 1-qubit gate .

    The 'euler' argument can be used to specify any of the 6 Euler
    decompositions: 'XYX', 'XZX', 'YXY', 'YZY', 'ZXZ', 'ZYZ' (Default)
    """
    if euler == "ZYZ":
        return zyz_decomposition(gate)

    euler_trans = {
        "XYX": Y(0) ** 0.5,
        "XZX": Rn(+2 * np.pi / 3, np.sqrt(1 / 3), np.sqrt(1 / 3), np.sqrt(1 / 3), 0),
        "YXY": Rn(-2 * np.pi / 3, np.sqrt(1 / 3), np.sqrt(1 / 3), np.sqrt(1 / 3), 0),
        "YZY": Rn(np.pi, 0, np.sqrt(1 / 2), np.sqrt(1 / 2), 0),
        "ZXZ": S_H(0),
        "ZYZ": I(0),
    }

    q0 = gate.qubits[0]
    trans = euler_trans[euler].on(q0)
    gate = Circuit([trans, gate, trans.H]).asgate()
    zyz = zyz_decomposition(gate)
    params = [elem.param("t") for elem in zyz]  # type: ignore
    return euler_circuit(params[0], params[1], params[2], q0, euler)


# FIXME: Return 2 sub circuits?
def kronecker_decomposition(gate: Gate, euler: str = "ZYZ") -> Circuit:
    """
    Decompose a 2-qubit unitary composed of two 1-qubit local gates.

    Uses the "Nearest Kronecker Product" algorithm. Will give erratic
    results if the gate is not the direct product of two 1-qubit gates.
    """
    # An alternative approach would be to take partial traces, but
    # this approach appears to be more robust.

    if gate.qubit_nb != 2:
        raise ValueError("Expected 2-qubit gate")

    U = gate.asoperator()
    rank = 2 ** gate.qubit_nb
    U /= np.linalg.det(U) ** (1 / rank)

    R = U.reshape(2, 2, 2, 2)
    R = R.transpose(0, 2, 1, 3)
    R = R.reshape(4, 4)

    u, s, vh = np.linalg.svd(R)
    A = np.sqrt(s[0]) * u[:, 0].reshape(2, 2)
    B = np.sqrt(s[0]) * vh[0, :].reshape(2, 2)
    # print(s)
    # A = (u[:, 0]).reshape(2, 2)
    # B = (vh[0, :]).reshape(2, 2)

    q0, q1 = gate.qubits
    g0 = Unitary(A, [q0])
    g1 = Unitary(B, [q1])

    if not gates_close(gate, Circuit([g0, g1]).asgate()):
        raise ValueError("Gate cannot be decomposed into two 1-qubit gates")

    circ = Circuit()
    circ += euler_decomposition(g0, euler)
    circ += euler_decomposition(g1, euler)

    assert gates_close(gate, circ.asgate())  # Sanity check

    return circ


def canonical_coords(gate: Gate) -> Sequence[float]:
    """Returns the canonical coordinates of a 2-qubit gate"""
    circ = canonical_decomposition(gate)
    gate = circ[1]  # type: ignore
    params = [gate.param(key) for key in ("tx", "ty", "tz")]
    return params  # type: ignore


# FIXME: Does not return a circuit of 5 gates.
# Instead return locals_before, canonical, locals_after
def canonical_decomposition(gate: Gate, euler: str = "ZYZ") -> Circuit:
    """Decompose a 2-qubit gate by removing local 1-qubit gates to leave
    the non-local canonical two-qubit gate. [1]_ [2]_ [3]_ [4]_

    Returns: A Circuit of 3 operations: a circuit of initial 1-qubit gates;
    a canonical gate, with coordinates in the Weyl chamber; and a final
    circuit contains the final 1-qubits gates.

    The canonical coordinates can be found in circ.elements[2].params

    More or less follows the algorithm outlined in [2]_.

    .. [1] A geometric theory of non-local two-qubit operations, J. Zhang,
        J. Vala, K. B. Whaley, S. Sastry quant-ph/0291120
    .. [2] An analytical decomposition protocol for optimal implementation of
        two-qubit entangling gates. M. Blaauboer, R.L. de Visser,
        cond-mat/0609750
    .. [3] Metric structure of two-qubit gates, perfect entangles and quantum
        control, P. Watts, M. O'Conner, J. Vala, Entropy (2013)
    .. [4] Constructive Quantum Shannon Decomposition from Cartan Involutions
        B. Drury, P. Love, arXiv:0806.4015
    """

    # Implementation note: The canonical decomposition is easy. Constraining
    # canonical coordinates to the Weyl chamber is easy. But doing the
    # canonical decomposition with the canonical gate in the Weyl chamber
    # proved to be surprisingly tricky.

    # Unitary transform to Magic Basis of Bell states
    Q = np.asarray(
        [[1, 0, 0, 1j], [0, 1j, 1, 0], [0, 1j, -1, 0], [1, 0, 0, -1j]]
    ) / np.sqrt(2)
    Q_H = Q.conj().T

    if gate.qubit_nb != 2:
        raise ValueError("Expected 2-qubit gate")

    q0, q1 = gate.qubits

    U = gate.asoperator()
    rank = 2 ** gate.qubit_nb
    U /= np.linalg.det(U) ** (1 / rank)  # U is in SU(4) so det U = 1

    U_mb = Q_H @ U @ Q  # Transform gate to Magic Basis [1, (eq. 17, 18)]
    M = U_mb.transpose() @ U_mb  # Construct M matrix [1, (eq. 22)]

    # Diagonalize symmetric complex matrix
    eigvals, eigvecs = _eig_complex_symmetric(M)

    lambdas = np.sqrt(eigvals)  # Eigenvalues of F
    # Lambdas only fixed up to a sign. So make sure det F = 1 as it should
    det_F = np.prod(lambdas)
    if det_F.real < 0:
        lambdas[0] *= -1

    coords, signs, perm = _constrain_to_weyl(lambdas)

    # Construct local and canonical gates in magic basis
    lambdas = (lambdas * signs)[perm]
    O2 = (np.diag(signs) @ eigvecs.transpose())[perm]
    F = np.diag(lambdas)
    O1 = U_mb @ O2.transpose() @ F.conj()

    # Sanity check: Make sure O1 and O2 are orthogonal
    assert np.allclose(np.eye(4), O2.transpose() @ O2)  # Sanity check
    assert np.allclose(np.eye(4), O1.transpose() @ O1)  # Sanity check

    # Sometimes O1 & O2 end up with det = -1, instead of +1 as they should.
    # We can commute a diagonal matrix through F to fix this up.
    neg = np.diag([-1, 1, 1, 1])
    if np.linalg.det(O2).real < 0:
        O2 = neg @ O2
        O1 = O1 @ neg

    # Transform gates back from magic basis
    K1 = Q @ O1 @ Q_H
    A = Q @ F @ Q_H
    K2 = Q @ O2 @ Q_H

    assert gates_close(Unitary(U, [0, 1]), Unitary(K1 @ A @ K2, [0, 1]))  # Sanity check
    canon = Can(coords[0], coords[1], coords[2], q0, q1)

    # Sanity check
    assert gates_close(Unitary(A, gate.qubits), canon, atol=1e-4)

    # Decompose local gates into the two component 1-qubit gates
    gateK1 = Unitary(K1, gate.qubits)
    circK1 = kronecker_decomposition(gateK1, euler)
    assert gates_close(gateK1, circK1.asgate())  # Sanity check

    gateK2 = Unitary(K2, gate.qubits)
    circK2 = kronecker_decomposition(gateK2, euler)
    assert gates_close(gateK2, circK2.asgate())  # Sanity check

    # Build and return circuit
    # circ = Circuit()
    # circ += circK2
    # circ += canon
    # circ += circK1

    circ = Circuit([circK2, canon, circK1])

    return circ


def _eig_complex_symmetric(M: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    """Diagonalize a complex symmetric  matrix. The eigenvalues are
    complex, and the eigenvectors form an orthogonal matrix.

    Returns:
        eigenvalues, eigenvectors
    """
    if not np.allclose(M, M.transpose()):
        raise np.linalg.LinAlgError("Not a symmetric matrix")

    # The matrix of eigenvectors should be orthogonal.
    # But the standard 'eig' method will fail to return an orthogonal
    # eigenvector matrix when the eigenvalues are degenerate. However,
    # both the real and
    # imaginary part of M must be symmetric with the same orthogonal
    # matrix of eigenvectors. But either the real or imaginary part could
    # vanish. So we use a randomized algorithm where we diagonalize a
    # random linear combination of real and imaginary parts to find the
    # eigenvectors, taking advantage of the 'eigh' subroutine for
    # diagonalizing symmetric matrices.
    # This can fail if we're very unlucky with our random coefficient, so we
    # give the algorithm a few chances to succeed.

    # Empirically, never seems to fail on randomly sampled complex
    # symmetric 4x4 matrices.
    # If failure rate is less than 1 in a million, then 16 rounds
    # will have overall failure rate less than 1 in a googol.
    # However, cannot (yet) guarantee that there aren't special cases
    # which have much higher failure rates.

    # GEC 2018

    max_attempts = 16
    for _ in range(max_attempts):
        c = np.random.uniform(0, 1)
        matrix = c * M.real + (1 - c) * M.imag
        _, eigvecs = np.linalg.eigh(matrix)
        eigvecs = np.array(eigvecs, dtype=complex)
        eigvals = np.diag(eigvecs.transpose() @ M @ eigvecs)

        # Finish if we got a correct answer.
        reconstructed = eigvecs @ np.diag(eigvals) @ eigvecs.transpose()
        if np.allclose(M, reconstructed):
            return eigvals, eigvecs

    # Should never happen. Hopefully.
    raise np.linalg.LinAlgError(
        "Cannot diagonalize complex symmetric matrix."
    )  # pragma: no cover


def _lambdas_to_coords(lambdas: Sequence[float]) -> np.ndarray:
    # [2, eq.11], but using [1]s coordinates.
    l1, l2, _, l4 = lambdas
    c1 = np.real(1j * np.log(l1 * l2))
    c2 = np.real(1j * np.log(l2 * l4))
    c3 = np.real(1j * np.log(l1 * l4))
    coords = np.asarray((c1, c2, c3)) / np.pi

    coords[np.abs(coords - 1) < ATOL] = -1
    if all(coords < 0):
        coords += 1

    # If we're close to the boundary, floating point errors can conspire
    # to make it seem that we're never on the inside
    # Fix: If near boundary, reset to boundary

    # Left
    if np.abs(coords[0] - coords[1]) < ATOL:
        coords[1] = coords[0]

    # Front
    if np.abs(coords[1] - coords[2]) < ATOL:
        coords[2] = coords[1]

    # Right
    if np.abs(coords[0] - coords[1] - 1 / 2) < ATOL:
        coords[1] = coords[0] - 1 / 2

    # Base
    coords[np.abs(coords) < ATOL] = 0

    return coords


def _constrain_to_weyl(
    lambdas: Sequence[float],
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:

    for permutation in itertools.permutations(range(4)):
        for signs in ([1, 1, 1, 1], [1, 1, -1, -1], [-1, 1, -1, 1], [1, -1, -1, 1]):
            signed_lambdas = lambdas * np.asarray(signs)
            perm = list(permutation)
            lambdas_perm = signed_lambdas[perm]

            coords = _lambdas_to_coords(lambdas_perm)

            if _in_weyl(*coords):
                return coords, np.asarray(signs), perm

    # Should never get here
    assert False  # pragma: no cover
    return None, None, None  # pragma: no cover


def _in_weyl(tx: float, ty: float, tz: float) -> bool:
    # Note 'tz>0' in second term. This takes care of symmetry across base
    # when tz==0
    return (1 / 2 >= tx >= ty >= tz >= 0) or (1 / 2 >= (1 - tx) >= ty >= tz > 0)


def cnot_decomposition(gate: Gate) -> Circuit:
    """Decompose any 2-qubit gate into a circuit of three CNot gates.

    Ref:
        Optimal Quantum Circuits for General Two-Qubit Gates, Vatan & Williams
        (2004) (quant-ph/0308006) Fig. 6
    """
    q0, q1 = gate.qubits

    gate_dek = canonical_decomposition(gate)
    gate_before = cast(Circuit, gate_dek[0])
    gate_can = cast(Can, gate_dek[1])
    gate_after = cast(Circuit, gate_dek[2])

    cnot_circ = Circuit(translate_can_to_cnot(gate_can))
    assert gates_close(gate_can, cnot_circ.asgate())
    circ = gate_before + cnot_circ + gate_after

    return circ


def b_decomposition(gate: Gate) -> Circuit:
    """Decompose any 2-qubit gate into a sandwich of two B gates.

    Refs:
        Minimum construction of two-qubit quantum operations
        https://arxiv.org/abs/quant-ph/0312193
    """
    # Implementation Note
    # The analytic construction of a given canonical gate, up to local gates, is
    # given in quant-ph/0312193. But we don't know how to get the local
    # gates analytically. (Although, it looks like it could be done...)
    # Instead we follow the tactic used in cirq. We take a canonical
    # decomposition of the original gate, construct a B gate sandwich with the
    # same canonical coordinates, then canonically decompose the B gate
    # sandwich to figure out the appropriate local gates.
    # Kudos: Cirq, @Strilanc
    # Ref: https://github.com/quantumlib/Cirq/pull/2574/files

    q0, q1 = gate.qubits

    gate_dek = canonical_decomposition(gate)
    gate_before = cast(Circuit, gate_dek[0])
    gate_can = gate_dek[1]
    gate_after = cast(Circuit, gate_dek[2])

    tx, ty, tz = gate_can.params
    c0 = np.sin(0.5 * ty * np.pi) ** 2 * np.cos(0.5 * tz * np.pi) ** 2
    c0 = max(0.0, min(0.5, c0))
    c1 = (np.cos(ty * np.pi) * np.cos(tz * np.pi)) / (1 - 2 * c0)
    c1 = max(0.0, min(1.0, c1))

    sy = (1 / np.pi) * np.arccos(1 - 4 * c0)
    sz = -(1 / np.pi) * np.arcsin(np.sqrt(c1))

    bcirc = Circuit(
        [
            B(q0, q1),
            YPow(-tx, q0),
            ZPow(sz, q1),
            YPow(sy, q1),
            ZPow(sz, q1),
            B(q0, q1),
        ]
    )

    bcirc_dek = canonical_decomposition(bcirc.asgate())
    bcirc_before = cast(Circuit, bcirc_dek[0])
    bcirc_after = cast(Circuit, bcirc_dek[2])

    circ = gate_before + bcirc_before.H + bcirc + bcirc_after.H + gate_after

    return circ


# TODO: Should be able to replace _lambdas_to_coords with this method,
# which should be faster
# TODO: Better name?
# DOCME
def convert_can_to_weyl(gate: Can, euler: str = "ZYZ") -> Circuit:
    """Move a Canonical gate with arbitrary coordinates to a canonical gate
    in the Weyl chamber, plus local 1-qubit gates.

    Args:
        gate:  A canonical (Can) gate
        euler: The Euler decomposition to use when simplifying the local gates
                If None, do not simplify.
    Returns:
        A circuit of 5 elements --
            circuit of gates acting on the first qubit
            circuit of gates acting on the second qubit
            canonical gate in the Weyl chamber
            circuit of gates acting on the first qubit
            circuit of gates acting on the second qubit
    """

    tx, ty, tz = gate.params
    q0, q1 = gate.qubits

    # Local gates to put before the canonical Weyl gate
    before_q0: List[Gate] = [I(q0)]
    before_q1: List[Gate] = [I(q1)]

    # Local gates that come after the canonical Weyl gate, in reverse
    # (complex conjugate) order.
    after_q0_H: List[Gate] = [I(q0)]
    after_q1_H: List[Gate] = [I(q1)]

    def tx_minus_1() -> None:
        before_q0.append(Y(q0))
        before_q1.append(Y(q1))
        after_q0_H.append(Z(q0))
        after_q1_H.append(Z(q1))

    def ty_minus_1() -> None:
        before_q0.append(X(q0))
        before_q1.append(X(q1))
        after_q0_H.append(Z(q0))
        after_q1_H.append(Z(q1))

    def tz_minus_1() -> None:
        before_q0.append(X(q0))
        before_q1.append(X(q1))
        after_q0_H.append(Y(q0))
        after_q1_H.append(Y(q1))

    def flip_tx_ty() -> None:
        before_q0.append(Z(q0))
        after_q0_H.append(Z(q0))

    def flip_tx_tz() -> None:
        before_q0.append(Y(q0))
        after_q0_H.append(Y(q0))

    def flip_ty_tz() -> None:
        before_q0.append(X(q0))
        after_q0_H.append(X(q0))

    def swap_tx_ty() -> None:
        before_q0.append(S(q0))
        before_q1.append(S(q1))
        after_q0_H.append(S(q0))
        after_q1_H.append(S(q1))

    def swap_tx_tz() -> None:
        before_q0.append(Y(q0) ** 0.5)
        before_q1.append(Y(q1) ** 0.5)
        after_q0_H.append(Y(q0) ** 0.5)
        after_q1_H.append(Y(q1) ** 0.5)

    def swap_ty_tz() -> None:
        before_q0.append(V(q0))
        before_q1.append(V(q1))
        after_q0_H.append(V(q0))
        after_q1_H.append(V(q1))

    tx = tx % 2
    ty = ty % 2
    tz = tz % 2

    if tx > 1.0:
        tx -= 1
        tx_minus_1()

    if ty > 1.0:
        ty -= 1
        ty_minus_1()

    if tz > 1.0:
        tz -= 1
        tz_minus_1()

    if tx > 0.5 and ty > 0.5:
        tx = 1 - tx
        ty = 1 - ty
        tx_minus_1()
        ty_minus_1()
        flip_tx_ty()

    if tx > 0.5 and tz > 0.5:
        tx = 1 - tx
        tz = 1 - tz
        tx_minus_1()
        tz_minus_1()
        flip_tx_tz()

    if ty > 0.5 and tz > 0.5:
        ty = 1 - ty
        tz = 1 - tz
        ty_minus_1()
        tz_minus_1()
        flip_ty_tz()

    if ty > tx:
        tx, ty = ty, tx
        swap_tx_ty()

    if tz > tx:
        tx, tz = tz, tx
        swap_tx_tz()

    if tz > ty:
        tz, ty = ty, tz
        swap_ty_tz()

    if tx > 0.5 and (1 - tx) < ty:
        tx, ty = ty, tx
        swap_tx_ty()
        tx = 1 - tx
        ty = 1 - ty
        tx_minus_1()
        ty_minus_1()
        flip_tx_ty()

    if ty > tx:  # pragma: no cover   # Never happens?
        tx, ty = ty, tx
        swap_tx_ty()

    if tz > tx:  # pragma: no cover   # Never happens?
        tx, tz = tz, tx
        swap_tx_tz()

    if tz > ty:
        tz, ty = ty, tz
        swap_ty_tz()

    # Insanity check
    assert _in_weyl(tx, ty, tz)

    circ_before_q0 = Circuit(before_q0)
    circ_before_q1 = Circuit(before_q1)
    circ_after_q0_H = Circuit(after_q0_H)
    circ_after_q1_H = Circuit(after_q1_H)
    if euler is not None:
        circ_before_q0 = euler_decomposition(circ_before_q0.asgate(), euler=euler)
        circ_before_q1 = euler_decomposition(circ_before_q1.asgate(), euler=euler)
        circ_after_q0_H = euler_decomposition(circ_after_q0_H.asgate(), euler=euler)
        circ_after_q1_H = euler_decomposition(circ_after_q1_H.asgate(), euler=euler)

    circ = Circuit()

    circ += circ_before_q0
    circ += circ_before_q1
    circ += Can(tx, ty, tz, q0, q1)
    circ += circ_after_q0_H.H
    circ += circ_after_q1_H.H

    return circ


# fin



# Copyright 2020-, Gavin E. Crooks and contributors
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.


import numpy as np

import quantumflow as qf


def test_ccnot() -> None:
    ket = qf.zero_state(3)
    ket = qf.CCNot(0, 1, 2).run(ket)
    assert ket.tensor[0, 0, 0] == 1.0

    ket = qf.X(1).run(ket)
    ket = qf.CCNot(0, 1, 2).run(ket)
    assert ket.tensor[0, 1, 0] == 1.0

    ket = qf.X(0).run(ket)
    ket = qf.CCNot(0, 1, 2).run(ket)
    assert ket.tensor[1, 1, 1] == 1.0


def test_cswap() -> None:
    ket = qf.zero_state(3)
    ket = qf.X(1).run(ket)
    ket = qf.CSwap(0, 1, 2).run(ket)
    assert ket.tensor[0, 1, 0] == 1.0

    ket = qf.X(0).run(ket)
    ket = qf.CSwap(0, 1, 2).run(ket)
    assert ket.tensor[1, 0, 1] == 1.0


def test_ccz() -> None:
    ket = qf.zero_state(3)
    ket = qf.X(0).run(ket)
    ket = qf.X(1).run(ket)
    ket = qf.H(2).run(ket)
    ket = qf.CCZ(0, 1, 2).run(ket)
    ket = qf.H(2).run(ket)
    qf.print_state(ket)
    assert np.isclose(ket.tensor[1, 1, 1], 1.0)

    gate0 = qf.CCZ()
    assert gate0.H is gate0


def test_deutsch() -> None:
    gate0 = qf.Deutsch(5 * np.pi / 2, 0, 1, 2)
    gate1 = qf.CCNot(0, 1, 2)
    assert qf.gates_close(gate0, gate1)


# fin



# Copyright 2020-, Gavin E. Crooks and contributors
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.


import random

import numpy as np
import pytest
from numpy import pi

import quantumflow as qf
from quantumflow import tensors

from .config_test import REPS


def test_asqutensor() -> None:
    arr = np.zeros(shape=(256,))
    tensor = qf.asqutensor(arr)
    assert tensor.shape == (2,) * 8


def test_asqutensor_flatten() -> None:
    arr = np.zeros(shape=(256,))
    tensor = qf.asqutensor(arr)

    arr = tensors.flatten(tensor, rank=1)
    assert arr.shape == (256,)

    arr = tensors.flatten(tensor, rank=2)
    assert arr.shape == (16,) * 2

    arr = tensors.flatten(tensor, rank=4)
    assert arr.shape == (4,) * 4

    arr = tensors.flatten(tensor, rank=8)
    assert arr.shape == (2,) * 8


def test_inner_product() -> None:
    # also tested via fubini_study_angle

    for _ in range(REPS):
        theta = random.uniform(-4 * pi, +4 * pi)

        hs = tensors.inner(qf.Rx(theta, 0).tensor, qf.Rx(theta, 0).tensor)
        print(f"Rx({theta}), hilbert_schmidt = {hs}")
        assert np.isclose(hs / 2, 1.0)

        hs = tensors.inner(qf.Rz(theta, 0).tensor, qf.Rz(theta, 0).tensor)
        print(f"Rz({theta}), hilbert_schmidt = {hs}")
        assert np.isclose(hs / 2, 1.0)

        hs = tensors.inner(qf.Ry(theta, 0).tensor, qf.Ry(theta, 0).tensor)
        print(f"Ry({theta}), hilbert_schmidt = {hs}")
        assert np.isclose(hs / 2, 1.0)

        hs = tensors.inner(qf.PSwap(theta, 0, 1).tensor, qf.PSwap(theta, 0, 1).tensor)
        print(f"PSwap({theta}), hilbert_schmidt = {hs}")
        assert np.isclose(hs / 4, 1.0)

    with pytest.raises(ValueError):
        tensors.inner(qf.zero_state(0).tensor, qf.X(0).tensor)

    with pytest.raises(ValueError):
        tensors.inner(qf.CNot(0, 1).tensor, qf.X(0).tensor)


# fin



# Copyright 2019-, Gavin E. Crooks and the QuantumFlow contributors
# Copyright 2016-2018, Rigetti Computing
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.


"""
Unit tests for quantumflow.paulialgebra
"""

from itertools import product

import numpy as np
import pytest
import scipy.linalg

import quantumflow as qf
from quantumflow.paulialgebra import PAULI_OPS, sI, sX, sY, sZ


def test_term() -> None:
    x = qf.Pauli.term([0], "X", -1)
    assert x.terms == (((0,), "X", -1),)

    x = qf.Pauli.term([1, 0, 5], "XYZ", 2j)
    assert x.terms == (((0, 1, 5), "YXZ", (0 + 2j)),)

    with pytest.raises(ValueError):
        qf.Pauli.term([1, 0, 5], "BYZ", 2j)


def test_pauli_str() -> None:
    x = qf.Pauli.term([2], "X", -1)
    assert str(x) == "- sX(2)"
    assert repr(x) == "Pauli((((2,), 'X', -1),))"


def test_pauli_sigma() -> None:
    assert qf.Pauli.term([0], "X", 1) == qf.Pauli.sigma(0, "X")
    assert qf.Pauli.term([1], "Y", 2) == qf.Pauli.sigma(1, "Y", 2)
    assert qf.Pauli.term([2], "Z") == qf.Pauli.sigma(2, "Z")
    # assert qf.Pauli.term([3], 'I', 1) == qf.Pauli.sigma('I', 3) # FIXME


def test_sigmas() -> None:
    assert qf.sX(0).terms == (((0,), "X", 1),)
    assert qf.sY(1).terms == (((1,), "Y", 1),)
    assert qf.sZ(2).terms == (((2,), "Z", 1),)


def test_sum() -> None:
    x = qf.sX(1)
    y = qf.sY(2, 2.0)
    s = qf.pauli_sum(x, y)
    assert s == qf.Pauli(([1], "X", 1.0), ([2], "Y", 2.0))

    s2 = qf.pauli_sum(x, x)
    assert s2 == qf.Pauli(([1], "X", 2))

    s3 = qf.pauli_sum(x, x, y)
    s4 = qf.pauli_sum(y, x, x)

    assert s3 == s4
    assert s3 == qf.Pauli(([1], "X", 2), ([2], "Y", 2))

    qf.pauli_sum(x, x, x)


def test_add() -> None:
    x = sX(1)
    y = sY(2, 2.0)
    s = y + x
    assert s == qf.Pauli(([1], "X", 1), ([2], "Y", 2))


def test_sub() -> None:
    x = sX(1)
    y = sY(2, 2.0)
    s = x - y
    assert s == qf.Pauli(([1], "X", 1), ([2], "Y", -2))

    s = 2 - y
    assert str(s) == "+(2+0j) +(-2+0j) sY(2)"


def test_cmp() -> None:
    x = sX(1)
    y = sY(2, 2.0)
    assert x < y
    assert x <= y
    assert x <= x
    assert x == x
    assert y > x
    assert y >= y

    with pytest.raises(TypeError):
        "foo" > y

    with pytest.raises(TypeError):
        "foo" >= y

    with pytest.raises(TypeError):
        "foo" < y

    with pytest.raises(TypeError):
        "foo" <= y

    assert not 2 == y


def test_hash() -> None:
    x = sX(5)
    d = {x: 4}
    assert d[x] == 4


def test_product() -> None:
    p = qf.pauli_product(sX(0), sY(0))
    assert p == sZ(0, 1j)

    p = qf.pauli_product(sY(0), sX(0))
    assert p == sZ(0, -1j)

    p = qf.pauli_product(sX(0), sY(1))
    assert p == qf.Pauli.term([0, 1], "XY", 1)

    p = qf.pauli_product(sX(0), sY(1), sY(0))
    assert p == qf.Pauli.term([0, 1], "ZY", 1j)

    p = qf.pauli_product(sY(0), sX(0), sY(1))
    assert p == qf.Pauli.term([0, 1], "ZY", -1j)


def test_mul() -> None:
    # TODO CHECK ALL PAULI MULTIPLICATIONS HERE
    assert sX(0) * sY(0) == sZ(0, 1j)


def test_scalar() -> None:
    a = qf.Pauli.scalar(1.0)
    assert a.is_scalar()
    assert a.is_identity()

    b = qf.Pauli.scalar(2.0)
    assert b + b == qf.Pauli.scalar(4.0)
    assert b * b == qf.Pauli.scalar(4.0)

    assert -b == qf.Pauli.scalar(-2.0)
    assert +b == qf.Pauli.scalar(2.0)

    assert b * 3 == qf.Pauli.scalar(6.0)
    assert 3 * b == qf.Pauli.scalar(6.0)

    assert sX(0) * 2 == sX(0, 2)

    x = sX(0) + sY(1)
    assert not x.is_scalar()
    assert not x.is_identity()

    c = sX(0) * sY(1)
    assert not c.is_scalar()
    assert not c.is_identity()


def test_zero() -> None:
    z = qf.Pauli.scalar(0.0)
    assert z.is_zero()
    assert z.is_scalar()

    z2 = qf.Pauli.zero()
    assert z == z2

    assert sX(0) - sX(0) == qf.Pauli.zero()


def test_merge_sum() -> None:
    p = qf.pauli_sum(qf.Pauli.term([1], "Y", 3), qf.Pauli.term([1], "Y", 2))

    assert len(p) == 1
    assert p.terms[0][2] == 5


def test_power() -> None:
    p = sX(0) + sY(1) + qf.Pauli.term([2, 3], "XY")

    assert p ** 0 == qf.Pauli.identity()
    assert p ** 1 == p
    assert p * p == p ** 2
    assert p * p * p == p ** 3
    assert p * p * p * p * p * p * p * p * p * p == p ** 10

    with pytest.raises(ValueError):
        p ** -1

    p ** 400


def test_simplify() -> None:
    t1 = sZ(0) * sZ(1)
    t2 = sZ(0) * sZ(1)
    assert (t1 + t2) == 2 * sZ(0) * sZ(1)


def test_dont_simplify() -> None:
    t1 = sZ(0) * sZ(1)
    t2 = sZ(2) * sZ(3)
    assert (t1 + t2) != 2 * sZ(0) * sZ(1)


def test_zero_term() -> None:
    qubit_id = 0
    coefficient = 10
    ps = sI(qubit_id) + sX(qubit_id)
    assert coefficient * qf.Pauli.zero() == qf.Pauli.zero()
    assert qf.Pauli.zero() * coefficient == qf.Pauli.zero()
    assert qf.Pauli.zero() * qf.Pauli.identity() == qf.Pauli.zero()
    assert qf.Pauli.zero() + qf.Pauli.identity() == qf.Pauli.identity()
    assert qf.Pauli.zero() + ps == ps
    assert ps + qf.Pauli.zero() == ps


def test_neg() -> None:
    ps = sY(0) - sX(0)
    ps -= sZ(1)
    ps = ps - 3
    ps = 3 + ps


def test_paulisum_iteration() -> None:
    term_list = [sX(2), sZ(4)]
    ps = sX(2) + sZ(4)
    for ii, term in enumerate(ps):
        assert term_list[ii].terms[0] == term


def test_check_commutation() -> None:
    term1 = sX(0) * sX(1)
    term2 = sY(0) * sY(1)
    term3 = sY(0) * sZ(2)

    assert qf.paulis_commute(term2, term3)
    assert qf.paulis_commute(term2, term3)
    assert not qf.paulis_commute(term1, term3)


def test_commuting_sets() -> None:
    term1 = sX(0) * sX(1)
    term2 = sY(0) * sY(1)
    term3 = sY(0) * sZ(2)
    ps = term1 + term2 + term3
    pcs = qf.pauli_commuting_sets(ps)
    assert len(pcs) == 2

    pcs = qf.pauli_commuting_sets(term1)
    assert len(pcs) == 1


def test_get_qubits() -> None:
    term = sZ(0) * sX(1)
    assert term.qubits == (0, 1)
    term += sY(10) + sX(2) + sY(2)
    assert term.qubits == (0, 1, 2, 10)

    sum_term = qf.Pauli.term([0], "X", 0.5) + 0.5j * qf.Pauli.term(
        [10], "Y"
    ) * qf.Pauli.term([0], "Y", 0.5j)
    assert sum_term.qubits == (0, 10)


def test_check_commutation_rigorous() -> None:
    # more rigorous test.  Get all operators in Pauli group
    N = 3
    qubits = list(range(N))
    ps = [qf.Pauli.term(qubits, "".join(ops)) for ops in product(PAULI_OPS, repeat=N)]

    commuting = []
    non_commuting = []
    for left, right in product(ps, ps):
        if left * right == right * left:
            commuting.append((left, right))
        else:
            non_commuting.append((left, right))

    # now that we have our sets let's check against our code.
    for left, right in non_commuting:
        assert not qf.paulis_commute(left, right)

    for left, right in commuting:
        assert qf.paulis_commute(left, right)


def test_isclose() -> None:
    x = sX(1)
    y = sY(2, 2.0)
    y2 = sY(2, 2.00000001)

    assert qf.paulis_close(x, x)
    assert not qf.paulis_close(x, y)
    assert qf.paulis_close(y, y2)


def test_run() -> None:
    x = sX(1)
    y = sY(2, 1.2)
    s = x + y

    ket0 = qf.zero_state(3)
    _ = s.run(ket0)


def test_pauli_decompose_hermitian() -> None:
    gate = qf.X(0)
    H = gate.asoperator()
    pl = qf.pauli_decompose_hermitian(H)
    assert np.allclose(pl.asoperator(), H)

    gate = qf.X(0)
    op = gate.asoperator()
    H = -scipy.linalg.logm(op) / 1.0j
    pl = qf.pauli_decompose_hermitian(H)
    assert np.allclose(pl.asoperator(), H)

    N = 4
    gate2 = qf.RandomGate(range(N))
    op = gate2.asoperator()
    H = -scipy.linalg.logm(op) / 1.0j
    pl = qf.pauli_decompose_hermitian(H)
    assert np.allclose(pl.asoperator(), H)

    op = np.ones(shape=[2, 2, 2])
    with pytest.raises(ValueError):
        qf.pauli_decompose_hermitian(op)

    op = np.ones(shape=[4, 4])
    op[0, 1] = 10000
    with pytest.raises(ValueError):
        qf.pauli_decompose_hermitian(op)

    op = np.ones(shape=[3, 3])
    with pytest.raises(ValueError):
        qf.pauli_decompose_hermitian(op)


def test_pauli_rewire() -> None:
    term = sZ(0) * sX(1)
    term = term.on(5, 4)
    assert term.qubits == (4, 5)
    assert term == sZ(5) * sX(4)
    term = term.rewire({5: "a", 4: "b"})
    assert term.qubits == ("a", "b")
    assert term == sZ("a") * sX("b")

    with pytest.raises(ValueError):
        term.on(1, 2, 3, 4)


# fin



# Copyright 2019-, Gavin E. Crooks and the QuantumFlow contributors
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

"""
QuantumFlow: Translate, transform, and compile circuits.
"""

# TODO: Split into separate modules
# TODO: Put translations into alphabetic order so they can be found

from typing import (
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    Sequence,
    Tuple,
    Type,
    Union,
)

from . import var
from .circuits import Circuit
from .ops import Gate
from .stdgates import (
    CCZ,
    CH,
    CRZ,
    CU3,
    CV,
    CV_H,
    CY,
    CZ,
    ECP,
    RZZ,
    S_H,
    T_H,
    U2,
    U3,
    V_H,
    XX,
    XY,
    YY,
    ZZ,
    B,
    Barenco,
    Can,
    CCiX,
    CCNot,
    CCXPow,
    CNot,
    CNotPow,
    CPhase,
    CPhase00,
    CPhase01,
    CPhase10,
    CrossResonance,
    CSwap,
    CYPow,
    CZPow,
    Deutsch,
    Exch,
    FSim,
    FSwap,
    FSwapPow,
    Givens,
    H,
    HPow,
    I,
    ISwap,
    Ph,
    PhasedX,
    PhasedXPow,
    PhaseShift,
    PSwap,
    Rn,
    Rx,
    Ry,
    Rz,
    S,
    SqrtISwap,
    SqrtISwap_H,
    SqrtSwap,
    SqrtSwap_H,
    SqrtY,
    SqrtY_H,
    Swap,
    Sycamore,
    T,
    V,
    W,
    X,
    XPow,
    Y,
    YPow,
    Z,
    ZPow,
)

# Implementation note:
# __all__ is auto-magically created at end of module
#
# These translations are all analytic, so that we can use symbolic
# parameters. Numerical decompositions can be found in decompositions.py
#
# Translations return an Iterator over gates, rather than a Circuit, in part
# so that we can use type annotations to keep track of the source and resultant
# gates of each translation.


def translation_source_gate(trans: Callable) -> Type[Gate]:
    return trans.__annotations__["gate"]


def translation_target_gates(trans: Callable) -> Tuple[Type[Gate]]:

    # FIXME: Fails with KeyError if no return type annotation.
    # Add a more informative error message.
    ret = trans.__annotations__["return"].__args__[0]
    if hasattr(ret, "__args__"):  # Union
        gates = ret.__args__
    else:
        gates = (ret,)

    return gates


# TODO: Rename
def select_translators(
    target_gates: Iterable[Type[Gate]], translators: Iterable[Callable] = None
) -> List[Callable]:
    """Return a list of translations that will translate source gates to target
    gates.

    If no translations are specified, we use all QuantumFlow translations
    listed in qf.TRANSLATORS

    For example, to convert a circuit to use gates understood by QUIL:
    ::

        trans = qf.select_translators(qf.QUIL_GATES)
        circ = qf.transform(circ, trans)

    """
    # Warning: Black Voodoo magic. We use python's type annotations to figure
    # out the source gate and target gates of a translation.

    if translators is None:
        source_trans = set(TRANSLATORS.values())
    else:
        source_trans = set(translators)

    out_trans = []
    target_gates = set(target_gates)

    while source_trans:  # Loop until we run out of translations
        for trans in list(source_trans):
            from_gate = translation_source_gate(trans)
            to_gates = translation_target_gates(trans)
            if from_gate in target_gates:
                # If translation's source gates are already in targets
                # then we don't need this translation. Discard.
                source_trans.remove(trans)
                break

            if target_gates.issuperset(to_gates):
                # If target gate of translation are already in
                # target list, and source gate isn't, move
                # translation to output list and source gate to targets.
                target_gates.add(from_gate)
                out_trans.append(trans)
                source_trans.remove(trans)
                break
        else:
            # If we got here, none of the remaining translations can be
            # used. Break out of while and return results.
            break  # pragma: no cover

    return out_trans


def circuit_translate(
    circ: Circuit,
    translators: Sequence = None,
    targets: Sequence[Type[Gate]] = None,
    recurse: bool = True,
) -> Circuit:
    """Apply a collection of translations to each gate in a circuit.
    If recurse, then apply translations to output of translations
    until translationally invariant.
    """
    if translators is not None and targets is not None:
        raise ValueError("Specify either targets or translators, not both")

    gates = list(reversed(list(circ)))
    translated: List[Gate] = []

    if translators is None:
        if targets is None:
            targets = [Can, XPow, ZPow, I, Ph]
        translators = select_translators(targets)

    # Use type annotations to do dynamic dispatch
    gateclass_translation = {
        translation_source_gate(trans): trans for trans in translators
    }

    while gates:
        gate = gates.pop()
        if type(gate) in gateclass_translation:
            trans = gateclass_translation[type(gate)](gate)
            if recurse:
                gates.extend(reversed(list(trans)))
            else:
                translated.extend(trans)
        else:
            translated += gate

    return Circuit(translated)


# 1-qubit gates


def translate_x_to_tx(gate: X) -> Iterator[XPow]:
    """Translate X gate to XPow"""
    (q0,) = gate.qubits
    yield XPow(1, q0)


def translate_y_to_ty(gate: Y) -> Iterator[YPow]:
    """Translate Y gate to YPow"""
    (q0,) = gate.qubits
    yield YPow(1, q0)


def translate_z_to_tz(gate: Z) -> Iterator[ZPow]:
    """Translate Z gate to ZPow"""
    (q0,) = gate.qubits
    yield ZPow(1, q0)


def translate_s_to_tz(gate: S) -> Iterator[ZPow]:
    """Translate S gate to ZPow"""
    (q0,) = gate.qubits
    yield ZPow(0.5, q0)


def translate_t_to_tz(gate: T) -> Iterator[ZPow]:
    """Translate T gate to ZPow"""
    (q0,) = gate.qubits
    yield ZPow(0.25, q0)


def translate_invs_to_tz(gate: S_H) -> Iterator[ZPow]:
    """Translate S.H gate to ZPow"""
    (q0,) = gate.qubits
    yield ZPow(-0.5, q0)


def translate_invt_to_tz(gate: T_H) -> Iterator[ZPow]:
    """Translate inverse T gate to Rz (a quil standard gate)"""
    (q0,) = gate.qubits
    yield ZPow(-0.25, q0)


def translate_rx_to_tx(gate: Rx) -> Iterator[XPow]:
    """Translate Rx gate to XPow"""
    (q0,) = gate.qubits
    (theta,) = gate.params
    t = theta / var.PI
    yield XPow(t, q0)


def translate_ry_to_ty(gate: Ry) -> Iterator[YPow]:
    """Translate Ry gate to YPow"""
    (q0,) = gate.qubits
    (theta,) = gate.params
    t = theta / var.PI
    yield YPow(t, q0)


def translate_rz_to_tz(gate: Rz) -> Iterator[ZPow]:
    """Translate Rz gate to ZPow"""
    (q0,) = gate.qubits
    (theta,) = gate.params
    t = theta / var.PI
    yield ZPow(t, q0)


def translate_rn_to_rz_ry(gate: Rn) -> Iterator[Union[Rz, Ry]]:
    """Translate Rn Bloch rotation to Rz Ry Rz Ry Rz.

    Refs:
        http://www.vcpc.univie.ac.at/~ian/hotlist/qc/talks/bloch-sphere-rotations.pdf
    """
    (q0,) = gate.qubits
    theta, nx, ny, nz = gate.params

    ang_y = var.arccos(nz)
    ang_z = var.arctan2(ny, nx)

    yield Rz(-ang_z, q0)
    yield Ry(-ang_y, q0)
    yield Rz(theta, q0)
    yield Ry(ang_y, q0)
    yield Rz(ang_z, q0)


def translate_phase_to_rz(gate: PhaseShift) -> Iterator[Rz]:
    """Translate Phase gate to Rz (ignoring global phase)"""
    (q0,) = gate.qubits
    theta = gate.param("theta")
    yield Rz(theta, q0)


def translate_sqrty_to_ty(gate: SqrtY) -> Iterator[YPow]:
    """Translate sqrt-Y gate to YPow"""
    (q0,) = gate.qubits
    yield YPow(0.5, q0)


def translate_sqrty_h_to_ty(gate: SqrtY_H) -> Iterator[YPow]:
    """Translate sqrt-Y gate to YPow"""
    (q0,) = gate.qubits
    yield YPow(-0.5, q0)


def translate_tx_to_rx(gate: XPow) -> Iterator[Rx]:
    """Translate XPow gate to Rx"""
    (q0,) = gate.qubits
    theta = gate.param("t") * var.PI
    yield Rx(theta, q0)


def translate_ty_to_ry(gate: YPow) -> Iterator[Ry]:
    """Translate YPow gate to Ry"""
    (q0,) = gate.qubits
    theta = gate.param("t") * var.PI
    yield Ry(theta, q0)


def translate_tz_to_rz(gate: ZPow) -> Iterator[Rz]:
    """Translate ZPow gate to Rz"""
    (q0,) = gate.qubits
    theta = gate.param("t") * var.PI
    yield Rz(theta, q0)


def translate_ty_to_xzx(gate: YPow) -> Iterator[Union[XPow, ZPow]]:
    """Translate YPow gate to XPow and ZPow gates"""
    (q0,) = gate.qubits
    t = gate.param("t")
    yield XPow(0.5, q0)
    yield ZPow(t, q0)
    yield XPow(-0.5, q0)


def translate_tx_to_zyz(gate: XPow) -> Iterator[Union[YPow, S, S_H]]:
    """Translate XPow gate to S and YPow gates"""
    (q0,) = gate.qubits
    t = gate.param("t")
    yield S(q0)
    yield YPow(t, q0)
    yield S_H(q0)


def translate_tz_to_xyx(gate: ZPow) -> Iterator[Union[YPow, V, V_H]]:
    """Translate ZPow gate to V and YPow gates"""
    (q0,) = gate.qubits
    t = gate.param("t")
    yield V_H(q0)
    yield YPow(t, q0)
    yield V(q0)


def translate_phased_x_to_zxz(gate: PhasedX) -> Iterator[Union[X, ZPow]]:
    """Translate YPow gate to XPow and ZPow gates"""
    (q0,) = gate.qubits
    p = gate.param("p")
    yield ZPow(-p, q0)
    yield X(q0)
    yield ZPow(p, q0)


def translate_phased_tx_to_zxz(gate: PhasedXPow) -> Iterator[Union[XPow, ZPow]]:
    """Translate YPow gate to XPow and ZPow gates"""
    (q0,) = gate.qubits
    p, t = gate.params
    yield ZPow(-p, q0)
    yield XPow(t, q0)
    yield ZPow(p, q0)


def translate_v_to_tx(gate: V) -> Iterator[XPow]:
    """Translate V gate to XPow"""
    (q0,) = gate.qubits
    yield XPow(0.5, q0)


def translate_invv_to_tx(gate: V_H) -> Iterator[XPow]:
    """Translate V_H gate to XPow"""
    (q0,) = gate.qubits
    yield XPow(-0.5, q0)


def translate_th_to_tx(gate: HPow) -> Iterator[Union[XPow, H, S, T, S_H, T_H]]:
    """Translate powers of the Hadamard gate to XPow and YPow"""
    (q0,) = gate.qubits
    (t,) = gate.params

    yield S(q0)
    yield H(q0)
    yield T(q0)
    yield XPow(t, q0)
    yield T(q0).H
    yield H(q0)
    yield S(q0).H


def translate_ty_to_zxz(gate: YPow) -> Iterator[Union[XPow, S, S_H]]:
    """Translate YPow gate to ZPow and XPow gates"""
    (q0,) = gate.qubits
    t = gate.param("t")
    yield S_H(q0)
    yield XPow(t, q0)
    yield S(q0)


def translate_tx_to_zxzxz(gate: XPow) -> Iterator[Union[XPow, ZPow]]:
    """Convert an arbitrary power of a Pauli-X gate to Z and V gates"""
    (q0,) = gate.qubits
    t = gate.param("t")

    if var.isclose(t, 0.5) or var.isclose(t, -0.5):
        yield gate
        return

    yield ZPow(0.5, q0)
    yield XPow(0.5, q0)
    yield ZPow(t, q0)
    yield XPow(-0.5, q0)
    yield ZPow(-0.5, q0)


def translate_hadamard_to_zxz(gate: H) -> Iterator[Union[XPow, ZPow]]:
    """Convert a Hadamard gate to a circuit with ZPow and XPow gates."""
    (q0,) = gate.qubits
    yield ZPow(0.5, q0)
    yield XPow(0.5, q0)
    yield ZPow(0.5, q0)


def translate_u3_to_zyz(gate: U3) -> Iterator[Union[Rz, Ry]]:
    """Translate QASMs U3 gate to Rz and Ry"""
    (q0,) = gate.qubits
    theta, phi, lam = gate.params
    yield Rz(lam, q0)
    yield Ry(theta, q0)
    yield Rz(phi, q0)


def translate_u2_to_zyz(gate: U2) -> Iterator[Union[Rz, Ry]]:
    """Translate QASMs U2 gate to Rz and Ry"""
    (q0,) = gate.qubits
    phi, lam = gate.params
    yield Rz(lam, q0)
    yield Ry(var.PI / 2, q0)
    yield Rz(phi, q0)


def translate_tx_to_hzh(gate: XPow) -> Iterator[Union[H, ZPow]]:
    """Convert a XPow gate to a circuit with Hadamard and ZPow gates"""
    (q0,) = gate.qubits
    (t,) = gate.params
    yield H(q0)
    yield ZPow(t, q0)
    yield H(q0)


# 2-qubit gates


def translate_b_to_can(gate: B) -> Iterator[Union[Can, Y, Z]]:
    """Translate B gate to Canonical gate"""
    q0, q1 = gate.qubits
    yield Z(q0)
    yield Y(q1)
    yield Can(1 / 2, 1 / 4, 0, q0, q1)
    yield Y(q0)
    yield Z(q1)


def translate_barenco_to_xx(gate: Barenco) -> Iterator[Union[XX, YPow, ZPow]]:
    """Translate a Barenco gate to XX plus local gates"""
    phi, alpha, theta = gate.params

    ct = theta / var.PI
    ca = alpha / var.PI
    cp = phi / var.PI

    q0, q1 = gate.qubits

    yield ZPow(-1 / 2 + cp, q0)
    yield YPow(1 / 2, q0)
    yield ZPow(-1, q0)
    yield ZPow(1 / 2 - cp, q1)
    yield YPow(1 / 2, q1)
    yield ZPow(3 / 2, q1)
    yield XX(ct, q0, q1)
    yield YPow(1 / 2, q0)
    yield ZPow(3 / 2 + ca - cp, q0)
    yield ZPow(-1 / 2, q1)
    yield YPow(1 / 2 - ct, q1)
    yield ZPow(-3 / 2 + cp, q1)


def translate_can_to_cnot(
    gate: Can,
) -> Iterator[Union[CNot, S, S_H, XPow, YPow, ZPow, V, Z]]:
    """Translate canonical gate to 3 CNotS, using the Kraus-Cirac decomposition.

    ::

        ───Can(tx,ty,tz)───     ───────X───Z^tz-1/2───●──────────────X───S⁺───
            │                          │              │              │
        ───Can(tx,ty,tz)───     ───S───●───Y^tx-1/2───X───Y^1/2-ty───●────────

    Ref:
        Vatan and Williams. Optimal quantum circuits for general two-qubit gates.
        Phys. Rev. A, 69:032315, 2004. quant-ph/0308006 :cite:`Vatan2004` Fig. 6

        B. Kraus and J. I. Cirac, Phys. Rev. A 63, 062309 (2001).
    """  # noqa: W291, E501
    # TODO: Other special cases

    # Note: sign flip on central ZPow, YPow, YPow because of differing sign
    # conventions for Canonical.
    tx, ty, tz = gate.params
    q0, q1 = gate.qubits

    if var.isclose(tz, 0.0):
        # If we know tz is close to zero we only need two CNot gates
        yield V(q0)
        yield Z(q0)
        yield V(q1)
        yield Z(q1)
        yield CNot(q0, q1)
        yield XPow(tx, q0)
        yield ZPow(ty, q1)
        yield CNot(q0, q1)
        yield V(q0)
        yield Z(q0)
        yield V(q1)
        yield Z(q1)
    else:
        yield S(q1)
        yield CNot(q1, q0)
        yield ZPow(tz - 0.5, q0)
        yield YPow(-0.5 + tx, q1)
        yield CNot(q0, q1)
        yield YPow(-ty + 0.5, q1)
        yield CNot(q1, q0)
        yield S_H(q0)


def translate_can_to_xx_yy_zz(gate: Can) -> Iterator[Union[XX, YY, ZZ]]:
    """Convert a canonical gate to a circuit with XX, YY, and ZZ gates."""
    tx, ty, tz = gate.params
    q0, q1 = gate.qubits

    if not var.isclose(tx, 0.0):
        yield XX(tx, q0, q1)
    if not var.isclose(ty, 0.0):
        yield YY(ty, q0, q1)
    if not var.isclose(tz, 0.0):
        yield ZZ(tz, q0, q1)


def translate_ch_to_cpt(gate: CH) -> Iterator[Union[CNot, S, T, S_H, T_H, H]]:
    """Decomposition of a controlled Hadamard-gate into the Clifford+T.
    ::

        ───●───     ───────────────●─────────────────
           │                       │
        ───H───     ───S───H───T───X───T⁺───H───S⁺───

    Ref:
        http://arxiv.org/abs/1206.0758v3, Figure 5(a)
    """
    # Kudos: Adapted from QuipperLib
    # https://www.mathstat.dal.ca/~selinger/quipper/doc/QuipperLib-GateDecompositions.html
    q0, q1 = gate.qubits
    yield S(q1)
    yield H(q1)
    yield T(q1)
    yield CNot(q0, q1)
    yield T_H(q1)
    yield H(q1)
    yield S_H(q1)


def translate_cnot_to_cz(gate: CNot) -> Iterator[Union[H, CZ]]:
    """Convert CNot gate to a CZ based circuit."""
    q0, q1 = gate.qubits
    yield H(q1)
    yield CZ(q0, q1)
    yield H(q1)


def translate_cnot_to_sqrtiswap(gate: CNot) -> Iterator[Union[SqrtISwap_H, X, S_H, H]]:
    """Translate an ECP gate to a square-root-iswap sandwich"""
    q0, q1 = gate.qubits

    # TODO: simplify 1-qubit gates
    yield H(q1)
    yield S_H(q0)
    yield S_H(q1)
    yield H(q0)
    yield H(q1)
    yield SqrtISwap(q0, q1).H
    yield X(q0)
    yield SqrtISwap(q0, q1).H
    yield X(q0)
    yield H(q0)


def translate_cnot_to_sqrtswap(gate: CNot) -> Iterator[Union[SqrtSwap, YPow, ZPow, Z]]:
    """Translate square-root swap to canonical"""
    # https://qipc2011.ethz.ch/uploads/Schoolpresentations/berghaus2011_DiVincenzo.pdf
    q0, q1 = gate.qubits

    yield Y(q1) ** 0.5
    yield SqrtSwap(q0, q1)
    yield Z(q0)
    yield SqrtSwap(q0, q1)
    yield Z(q0) ** -0.5
    yield Z(q1) ** -0.5
    yield Y(q1) ** -0.5


def translate_cnot_to_xx(gate: CNot) -> Iterator[Union[XX, H, S_H]]:
    """Convert CNot to XX gate"""
    # TODO: simplify 1-qubit gates
    q0, q1 = gate.qubits
    yield H(q0)
    yield XX(0.5, q0, q1)
    yield H(q0)
    yield H(q1)
    yield S(q0).H
    yield S(q1).H
    yield H(q1)


def translate_cy_to_cnot(gate: CY) -> Iterator[Union[CNot, S, S_H]]:
    """Translate CY to CNot (CX)"""
    q0, q1 = gate.qubits
    yield S_H(q1)
    yield CNot(q0, q1)
    yield S(q1)


def translate_cypow_to_cxpow(gate: CYPow) -> Iterator[Union[CNotPow, S, S_H]]:
    """Translate powers of CY to powers of CNot (CX)"""
    yield from gate.decompose()


def translate_cphase_to_zz(gate: CPhase) -> Iterator[Union[ZZ, ZPow]]:
    """Convert a CPhase gate to a ZZ based circuit."""
    t = -gate.param("theta") / (2 * var.PI)
    q0, q1 = gate.qubits
    yield ZZ(t, q0, q1)
    yield ZPow(-t, q0)
    yield ZPow(-t, q1)


def translate_cphase00_to_zz(gate: CPhase00) -> Iterator[Union[X, ZZ, ZPow]]:
    """Convert a CPhase00 gate to a ZZ based circuit."""
    t = -gate.param("theta") / (2 * var.PI)
    q0, q1 = gate.qubits
    yield X(q0)
    yield X(q1)
    yield ZZ(t, q0, q1)
    yield ZPow(-t, q0)
    yield ZPow(-t, q1)
    yield X(q0)
    yield X(q1)


def translate_cphase01_to_zz(gate: CPhase01) -> Iterator[Union[X, ZZ, ZPow]]:
    """Convert a CPhase01 gate to a ZZ based circuit."""
    t = -gate.param("theta") / (2 * var.PI)
    q0, q1 = gate.qubits
    yield X(q0)
    yield ZZ(t, q0, q1)
    yield ZPow(-t, q0)
    yield ZPow(-t, q1)
    yield X(q0)


def translate_cphase10_to_zz(gate: CPhase10) -> Iterator[Union[X, ZZ, ZPow]]:
    """Convert a CPhase10 gate to a ZZ based circuit."""
    t = -gate.param("theta") / (2 * var.PI)
    q0, q1 = gate.qubits

    yield X(q1)
    yield ZZ(t, q0, q1)
    yield ZPow(-t, q0)
    yield ZPow(-t, q1)
    yield X(q1)


def translate_cross_resonance_to_xx(
    gate: CrossResonance,
) -> Iterator[Union[XX, XPow, YPow, X]]:
    """Translate a cross resonance gate to an XX based circuit"""
    s, b, c = gate.params
    q0, q1 = gate.qubits

    t7 = (
        var.arccos(
            ((1 + b ** 2 * var.cos(var.PI * var.sqrt(1 + b ** 2) * s))) / (1 + b ** 2)
        )
        / var.PI
    )
    t4 = c * s
    t1 = (
        var.arccos(
            var.cos(0.5 * var.PI * var.sqrt(1 + b ** 2) * s) / var.cos(t7 * var.PI / 2)
        )
        / var.PI
    )

    a = var.sin(var.PI * var.sqrt(1 + b ** 2) * s / 2)
    t7 *= var.sign(a) * var.sign(b)
    t1 *= var.sign(a)

    yield XPow(t1, q0)
    yield YPow(1.5, q0)
    yield X(q0)
    yield XPow(t4, q1)
    yield XX(t7, q0, q1)
    yield YPow(1.5, q0)
    yield X(q0)
    yield XPow(t1, q0)


def translate_crz_to_cnot(gate: CRZ) -> Iterator[Union[CNot, PhaseShift, U3]]:
    """Translate QASM's CRZ gate to standard gates.

    Ref:
        https://arxiv.org/pdf/1707.03429.pdf
    """
    q0, q1 = gate.qubits
    (theta,) = gate.params

    yield PhaseShift(theta / 2, q1)
    yield CNot(q0, q1)
    yield PhaseShift(-theta / 2, q1)
    yield CNot(q0, q1)


def translate_cnotpow_to_zz(gate: CNotPow) -> Iterator[Union[ZZ, ZPow, H]]:
    """Convert a controlled X^t gate to a ZZ based circuit.
    ::

        ───●─────     ───────ZZ^-t/2───Z^t/2───────
           │       =          │
        ───X^t───     ───H───ZZ^-t/2───Z^t/2───H───

    """
    (t,) = gate.params
    q0, q1 = gate.qubits
    yield H(q1)
    yield ZZ(-t / 2, q0, q1)
    yield ZPow(t / 2, q0)
    yield ZPow(t / 2, q1)
    yield H(q1)


def translate_cz_to_zz(gate: CZ) -> Iterator[Union[ZZ, S_H]]:
    """Convert CZ gate to a ZZ based circuit."""
    q0, q1 = gate.qubits
    yield ZZ(0.5, q0, q1)
    yield S_H(q0)
    yield S_H(q1)


def translate_czpow_to_zz(gate: CZPow) -> Iterator[Union[ZZ, ZPow]]:
    """Convert a CZPow gate to a ZZ based circuit."""
    t = gate.param("t")
    q0, q1 = gate.qubits
    yield ZZ(-t / 2, q0, q1)
    yield ZPow(t / 2, q0)
    yield ZPow(t / 2, q1)


def translate_czpow_to_cphase(gate: CZPow) -> Iterator[CPhase]:
    """Convert a CZPow gate to CPhase."""
    theta = gate.param("t") * var.PI
    yield CPhase(theta, *gate.qubits)


def translate_cphase_to_czpow(gate: CPhase) -> Iterator[CZPow]:
    """Convert a CPhase gate to a CZPow."""
    (theta,) = gate.params
    t = theta / var.PI
    yield CZPow(t, *gate.qubits)


# TODO: cphase to fsim
# TODO: fsim specialize


def translate_cu3_to_cnot(gate: CU3) -> Iterator[Union[CNot, PhaseShift, U3]]:
    """Translate QASM's CU3 gate to standard gates"""
    # Kudos: Adapted from qiskit
    # https://github.com/Qiskit/qiskit-terra/blob/master/qiskit/extensions/standard/cu3.py
    q0, q1 = gate.qubits
    theta, phi, lam = gate.params

    yield PhaseShift((lam + phi) / 2, q0)
    yield PhaseShift((lam - phi) / 2, q1)
    yield CNot(q0, q1)
    yield U3(-theta / 2, 0, -(phi + lam) / 2, q1)
    yield CNot(q0, q1)
    yield U3(theta / 2, phi, 0, q1)


def translate_cv_to_cpt(gate: CV) -> Iterator[Union[CNot, T, T_H, H]]:
    """Decomposition of a controlled sqrt(X)-gate into the Clifford+T.
    ::

        ───●───     ───T───X───T⁺───X───────
           │               │        │
        ───V───     ───H───●───T────●───H───

    Ref:
        http://arxiv.org/abs/1206.0758v3, Figure 5(c)
    """
    q0, q1 = gate.qubits
    yield T(q0)
    yield H(q1)
    yield CNot(q1, q0)
    yield T_H(q0)
    yield T(q1)
    yield CNot(q1, q0)
    yield H(q1)


def translate_cvh_to_cpt(gate: CV_H) -> Iterator[Union[CNot, T, T_H, H]]:
    """Decomposition of a controlled sqrt(X)-gate into the Clifford+T.
    ::

        ───●────     ───────X───T────X───T⁺───
           │                │        │
        ───V⁺───     ───H───●───T⁺───●───H────

    Ref:
        http://arxiv.org/abs/1206.0758v3, Figure 5(c)
    """
    q0, q1 = gate.qubits
    yield H(q1)
    yield CNot(q1, q0)
    yield T_H(q1)
    yield T(q0)
    yield CNot(q1, q0)
    yield H(q1)
    yield T_H(q0)


def translate_ecp_to_can(gate: ECP) -> Iterator[Can]:
    """Translate an ECP gate to a Canonical gate"""
    yield Can(1 / 2, 1 / 4, 1 / 4, *gate.qubits)


def translate_ecp_to_sqrtiswap(gate: ECP) -> Iterator[Union[SqrtISwap_H, YPow, S, S_H]]:
    """Translate an ECP gate to a square-root-iswap sandwich"""
    q0, q1 = gate.qubits

    yield SqrtISwap(q0, q1).H
    yield S(q0)
    yield S(q1)
    yield YPow(0.5, q0)
    yield YPow(0.5, q1)
    yield SqrtISwap(q0, q1).H

    yield YPow(-0.5, q1)
    yield S_H(q1)

    yield YPow(-0.5, q0)
    yield S_H(q0)


def translate_exch_to_can(gate: Exch) -> Iterator[Can]:
    """Convert an exchange gate to a canonical based circuit"""
    q0, q1 = gate.qubits
    t = gate.param("t")
    yield Can(t, t, t, q0, q1)


def translate_exch_to_xy_zz(gate: Exch) -> Iterator[Union[XY, ZZ]]:
    """Convert an exchange gate to XY and ZZ gates"""
    q0, q1 = gate.qubits
    t = gate.param("t")
    yield XY(t, q0, q1)
    yield ZZ(t, q0, q1)


def translate_fsim_to_xy_cz(gate: FSim) -> Iterator[Union[XY, CZ]]:
    """Convert the Cirq's FSim  gate to a canonical gate"""
    q0, q1 = gate.qubits
    theta, phi = gate.params

    yield XY(theta / var.PI, q0, q1)
    yield CZ(q0, q1) ** (-phi / var.PI)


def translate_fswap(gate: FSwap) -> Iterator[Union[Swap, CZ]]:
    """Translate fSwap gate to Swap and CV"""
    q0, q1 = gate.qubits
    yield Swap(q0, q1)
    yield CZ(q0, q1)


def translate_fswappow(gate: FSwapPow) -> Iterator[Union[Exch, CZPow]]:
    """Translate fSwap gate to XY and CVPow"""
    q0, q1 = gate.qubits
    t = gate.param("t")
    yield Swap(q0, q1) ** t
    yield CZ(q0, q1) ** t


# TODO: Other givens deke
def translate_givens_to_xy(gate: Givens) -> Iterator[Union[XY, T, T_H]]:
    """Convert a Givens  gate to an XY gate"""
    q0, q1 = gate.qubits
    (theta,) = gate.params

    yield T_H(q0)
    yield T(q1)
    yield XY(theta / var.PI, q0, q1)
    yield T(q0)
    yield T_H(q1)


def translate_iswap_to_can(gate: ISwap) -> Iterator[Union[Can, X]]:
    """Convert ISwap gate to a canonical gate within the Weyl chamber."""
    q0, q1 = gate.qubits
    yield X(q0)
    yield Can(0.5, 0.5, 0, q0, q1)
    yield X(q1)


def translate_iswap_to_swap_cz(gate: ISwap) -> Iterator[Union[Swap, CZ, S]]:
    """Convert ISwap gate to a Swap, CZ based circuit."""
    q0, q1 = gate.qubits
    yield Swap(q0, q1)
    yield CZ(q0, q1)
    yield S(q0)
    yield S(q1)


def translate_iswap_to_sqrtiswap(gate: ISwap) -> Iterator[SqrtISwap]:
    """Translate iswap gate to square-root iswaps"""
    q0, q1 = gate.qubits
    yield SqrtISwap(q0, q1)
    yield SqrtISwap(q0, q1)


def translate_iswap_to_xy(gate: ISwap) -> Iterator[Union[XY]]:
    """Convert ISwap gate to a XY gate."""
    q0, q1 = gate.qubits
    yield XY(-0.5, q0, q1)


def translate_pswap_to_canonical(gate: PSwap) -> Iterator[Union[Can, Y]]:
    """Translate parametric Swap to a canonical circuit"""

    q0, q1 = gate.qubits
    (theta,) = gate.params
    t = 0.5 - theta / var.PI
    yield Y(q0)
    yield Can(0.5, 0.5, t, q0, q1)
    yield Y(q1)


def translate_rzz_to_cnot(gate: RZZ) -> Iterator[Union[CNot, PhaseShift, U3]]:
    """Translate QASM's RZZ gate to standard gates"""
    q0, q1 = gate.qubits
    (theta,) = gate.params
    yield CNot(q0, q1)
    yield PhaseShift(theta, q1)
    yield CNot(q0, q1)


def translate_sqrtiswap_to_sqrtiswap_h(
    gate: SqrtISwap,
) -> Iterator[Union[SqrtISwap_H, Z]]:
    """Translate square-root-iswap to its inverse."""
    q0, q1 = gate.qubits
    yield Z(q0)
    yield SqrtISwap_H(q0, q1)
    yield Z(q0)


def translate_sqrtiswap_h_to_can(gate: SqrtISwap_H) -> Iterator[Can]:
    """Translate square-root iswap to canonical"""
    yield Can(1 / 4, 1 / 4, 0, *gate.qubits)


def translate_sqrtiswap_h_to_sqrtiswap(
    gate: SqrtISwap_H,
) -> Iterator[Union[SqrtISwap, Z]]:
    """Translate square-root-iswap to it's inverse."""
    q0, q1 = gate.qubits
    yield Z(q0)
    yield SqrtISwap(q0, q1)
    yield Z(q0)


def translate_sqrtswap_to_can(gate: SqrtSwap) -> Iterator[Can]:
    """Translate square-root swap to canonical"""
    yield Can(1 / 4, 1 / 4, 1 / 4, *gate.qubits)


def translate_sqrtswap_h_to_can(gate: SqrtSwap_H) -> Iterator[Can]:
    """Translate inv. square-root swap to canonical"""
    yield Can(-1 / 4, -1 / 4, -1 / 4, *gate.qubits)


def translate_swap_to_cnot(gate: Swap) -> Iterator[CNot]:
    """Convert a Swap gate to a circuit with 3 CNots."""
    q0, q1 = gate.qubits
    yield CNot(q0, q1)
    yield CNot(q1, q0)
    yield CNot(q0, q1)


def translate_swap_to_ecp_sqrtiswap(
    gate: Swap,
) -> Iterator[Union[ECP, SqrtISwap_H, H, ZPow, YPow]]:
    """Translate a Swap gate to an  ECP -- square-root-iswap sandwich.

    An intermediate step in translating swap to 3 square-root-iswap's.
    """
    q0, q1 = gate.qubits

    yield ECP(q0, q1)

    yield H(q0)
    yield H(q1)

    yield SqrtISwap(q0, q1).H

    yield YPow(-1 / 2, q1)
    yield ZPow(+1, q1)

    yield YPow(-1 / 2, q0)
    yield ZPow(+1, q0)


def translate_swap_to_iswap_cz(gate: Swap) -> Iterator[Union[ISwap, CZ, S_H]]:
    """Convert ISwap gate to a Swap, CZ based circuit."""
    q0, q1 = gate.qubits
    yield S_H(q0)
    yield S_H(q1)
    yield CZ(q0, q1)
    yield ISwap(q0, q1)


def translate_sycamore_to_fsim(gate: Sycamore) -> Iterator[FSim]:
    """Convert a Sycamore gate to an FSim gate"""
    yield FSim(var.PI / 2, var.PI / 6, *gate.qubits)


def translate_w_to_ecp(gate: W) -> Iterator[Union[ECP, H, S, S_H, T, T_H]]:
    """Translate W gate to ECP."""
    # TODO: Cite self
    q0, q1 = gate.qubits
    yield T(q0).H
    yield H(q0)
    yield T(q1).H
    yield S(q1).H
    yield H(q1)
    yield ECP(q0, q1)
    yield H(q0)
    yield S(q0)
    yield T(q0)
    yield H(q1)
    yield T(q1)


def translate_w_to_cnot(gate: W) -> Iterator[Union[CNot, H, S, S_H, T, T_H]]:
    """Translate W gate to CNot."""
    # Kudos: Decomposition given in Quipper
    # (Dek in paper far from optimal)
    # https://www.mathstat.dal.ca/~selinger/quipper/doc/QuipperLib-GateDecompositions.html
    q0, q1 = gate.qubits

    yield CNot(q0, q1)
    yield S(q0).H
    yield H(q0)
    yield T(q0).H
    yield CNot(q1, q0)
    yield T(q0)
    yield H(q0)
    yield S(q0)
    yield CNot(q0, q1)


def translate_w_to_ch_cnot(gate: W) -> Iterator[Union[CNot, CH]]:
    """Translate W gate to controlled-Hadamard and CNot."""
    # From https://arxiv.org/pdf/1505.06552.pdf
    q0, q1 = gate.qubits

    yield CNot(q1, q0)
    yield CNot(q0, q1)
    yield CH(q0, q1)
    yield CNot(q0, q1)
    yield CNot(q1, q0)


def translate_xx_to_can(gate: XX) -> Iterator[Union[Can]]:
    """Convert an XX gate to a canonical circuit."""
    q0, q1 = gate.qubits
    t = gate.param("t")
    yield Can(t, 0, 0, q0, q1)


def translate_xx_to_zz(gate: XX) -> Iterator[Union[H, ZZ]]:
    """Convert an XX gate to a ZZ based circuit."""
    q0, q1 = gate.qubits
    t = gate.param("t")
    yield H(q0)
    yield H(q1)
    yield ZZ(t, q0, q1)
    yield H(q0)
    yield H(q1)


def translate_xy_to_can(gate: XY) -> Iterator[Can]:
    """Convert XY gate to a canonical gate."""
    q0, q1 = gate.qubits
    t = gate.param("t")
    yield Can(t, t, 0, q0, q1)


def translate_xy_to_sqrtiswap(
    gate: XY,
) -> Iterator[Union[Z, T, T_H, ZPow, SqrtISwap_H]]:
    """Translate an XY gate to sandwich of square-root iswaps"""

    q0, q1 = gate.qubits
    t = gate.param("t")

    yield Z(q0)
    yield T(q0)
    yield T_H(q1)
    yield SqrtISwap_H(q0, q1)
    yield ZPow(1 - t, q0)
    yield ZPow(t, q1)
    yield SqrtISwap_H(q0, q1)
    yield T_H(q0)
    yield T(q1)


def translate_yy_to_can(gate: YY) -> Iterator[Union[Can]]:
    """Convert an YY gate to a canonical circuit."""
    q0, q1 = gate.qubits
    t = gate.param("t")
    yield Can(0, t, 0, q0, q1)


def translate_yy_to_zz(gate: YY) -> Iterator[Union[XPow, ZZ]]:
    """Convert a YY gate to a ZZ based circuit."""
    q0, q1 = gate.qubits
    t = gate.param("t")
    yield XPow(0.5, q0)
    yield XPow(0.5, q1)
    yield ZZ(t, q0, q1)
    yield XPow(-0.5, q0)
    yield XPow(-0.5, q1)


def translate_zz_to_can(gate: ZZ) -> Iterator[Union[Can]]:
    """Convert an ZZ gate to a canonical circuit."""
    q0, q1 = gate.qubits
    t = gate.param("t")
    yield Can(0, 0, t, q0, q1)


def translate_zz_to_cnot(gate: ZZ) -> Iterator[Union[CNot, ZPow]]:
    """Convert a ZZ gate to a CNot based circuit"""
    q0, q1 = gate.qubits
    t = gate.param("t")
    yield CNot(q0, q1)
    yield ZPow(t, q1)
    yield CNot(q0, q1)


def translate_zz_to_xx(gate: ZZ) -> Iterator[Union[H, XX]]:
    """Convert an XX gate to a ZZ based circuit."""
    q0, q1 = gate.qubits
    t = gate.param("t")
    yield H(q0)
    yield H(q1)
    yield XX(t, q0, q1)
    yield H(q0)
    yield H(q1)


def translate_zz_to_yy(gate: ZZ) -> Iterator[Union[XPow, YY]]:
    """Convert a YY gate to a ZZ based circuit."""
    q0, q1 = gate.qubits
    t = gate.param("t")
    yield XPow(-0.5, q0)
    yield XPow(-0.5, q1)
    yield YY(t, q0, q1)
    yield XPow(0.5, q0)
    yield XPow(0.5, q1)


# 3-qubit gates


def translate_ccix_to_cnot(gate: CCiX) -> Iterator[Union[CNot, H, T, T_H]]:
    """Decompose doubly-controlled iX-gate to 4 CNots.

    ::

        ───●────     ───────────────●────────────────●────────────
           │                        │                │
        ───●────  =  ───────●───────┼────────●───────┼────────────
           │                │       │        │       │
        ───iX───     ───H───X───T───X───T⁺───X───T───X───T⁺───H───


    Refs:
         Nielsen and Chuang (Figure 4.9)

    """
    # Kudos: Adapted from Quipper
    # https://www.mathstat.dal.ca/~selinger/quipper/doc/QuipperLib-GateDecompositions.html
    q0, q1, q2 = gate.qubits

    yield H(q2)
    yield CNot(q1, q2)
    yield T(q2)
    yield CNot(q0, q2)
    yield T(q2).H
    yield CNot(q1, q2)
    yield T(q2)
    yield CNot(q0, q2)
    yield T(q2).H
    yield H(q2)


def translate_ccix_to_cnot_adjacent(gate: CCiX) -> Iterator[Union[CNot, H, T, T_H]]:
    """Decompose doubly-controlled iX-gate to 8 CNots, respecting adjacency.

    ::

        ───●────     ───────────X───T⁺───────X────T───X───────X───
           │                    │            │        │       │
        ───●────  =  ───────X───●───T────X───●────X───●───X───●───
           │                │            │        │       │
        ───iX───     ───H───●────────────●───T⁺───●───────●───H───

    Refs:
          http://arxiv.org/abs/1210.0974, Figure 10
    """
    # Kudos: Adapted from 1210.0974, via Quipper
    q0, q1, q2 = gate.qubits

    yield H(q2)
    yield CNot(q2, q1)
    yield CNot(q1, q0)
    yield T(q0).H
    yield T(q1)
    yield CNot(q2, q1)
    yield CNot(q1, q0)
    yield T(q0)
    yield T(q2).H
    yield CNot(q2, q1)
    yield CNot(q1, q0)
    yield CNot(q2, q1)
    yield CNot(q1, q0)
    yield H(q2)


def translate_ccnot_to_ccz(gate: CCNot) -> Iterator[Union[H, CCZ]]:
    """Convert CCNot (Toffoli) gate to CCZ using Hadamards
    ::

        ───●───     ───────●───────
           │               │
        ───●───  =  ───────●───────
           │               │
        ───X───     ───H───●───H───

    """
    q0, q1, q2 = gate.qubits

    yield H(q2)
    yield CCZ(q0, q1, q2)
    yield H(q2)


def translate_ccnot_to_cnot(gate: CCNot) -> Iterator[Union[H, T, T_H, CNot]]:
    """Standard decomposition of CCNot (Toffoli) gate into six CNot gates.
    ::

        ───●───     ────────────────●────────────────●───●───T────●───
           │                        │                │   │        │
        ───●───  =  ───────●────────┼───────●───T────┼───X───T⁺───X───
           │               │        │       │        │
        ───X───     ───H───X───T⁺───X───T───X───T⁺───X───T───H────────

    Refs:
        M. A. Nielsen and I. L. Chuang, Quantum Computation and Quantum
        Information, Cambridge University Press (2000).
    """
    q0, q1, q2 = gate.qubits

    yield H(q2)
    yield CNot(q1, q2)
    yield T_H(q2)
    yield CNot(q0, q2)
    yield T(q2)
    yield CNot(q1, q2)
    yield T_H(q2)
    yield CNot(q0, q2)
    yield T(q1)
    yield T(q2)
    yield H(q2)
    yield CNot(q0, q1)
    yield T(q0)
    yield T_H(q1)
    yield CNot(q0, q1)


def translate_ccnot_to_cnot_AMMR(gate: CCNot) -> Iterator[Union[H, T, T_H, CNot]]:
    """Depth 9 decomposition of CCNot (Toffoli) gate into 7 CNot gates.
    ::

        ───●───     ───T───X───────●────────●────────T⁺───●───X───
           │               │       │        │             │   │
        ───●───  =  ───T───●───X───┼───T⁺───X───T⁺───X────┼───●───
           │                   │   │                 │    │
        ───X───     ───H───T───●───X────────────T────●────X───H───


    Refs:
        M. Amy, D. Maslov, M. Mosca, and M. Roetteler, A meet-in-the-middle
        algorithm for fast synthesis of depth-optimal quantum circuits, IEEE
        Transactions on Computer-Aided Design of Integrated Circuits and
        Systems 32(6):818-830.  http://arxiv.org/abs/1206.0758.
    """
    # Kudos: Adapted from QuipperLib
    # https://www.mathstat.dal.ca/~selinger/quipper/doc/QuipperLib-GateDecompositions.html

    q0, q1, q2 = gate.qubits

    yield H(q2)
    yield T(q0)
    yield T(q1)
    yield T(q2)
    yield CNot(q1, q0)
    yield CNot(q2, q1)
    yield CNot(q0, q2)
    yield T_H(q1)
    yield T(q2)
    yield CNot(q0, q1)
    yield T_H(q0)
    yield T_H(q1)
    yield CNot(q2, q1)
    yield CNot(q0, q2)
    yield CNot(q1, q0)
    yield H(q2)


def translate_ccnot_to_cv(gate: CCNot) -> Iterator[Union[CV, CV_H, CNot]]:
    """Decomposition of CCNot (Toffoli) gate into 3 CV and 2 CNots.
    ::

        ───●───     ───────●────────●───●───
           │               │        │   │
        ───●───  =  ───●───X───●────X───┼───
           │           │       │        │
        ───X───     ───V───────V⁺───────V───

    Refs:
         Barenco, Adriano; Bennett, Charles H.; Cleve, Richard;
         DiVincenzo, David P.; Margolus, Norman; Shor, Peter; Sleator, Tycho;
         Smolin, John A.; Weinfurter, Harald (Nov 1995). "Elementary gates for
         quantum computation". Physical Review A. American Physical Society. 52
         (5): 3457–3467. arXiv:quant-ph/9503016
    """

    q0, q1, q2 = gate.qubits
    yield CV(q1, q2)
    yield CNot(q0, q1)
    yield CV_H(q1, q2)
    yield CNot(q0, q1)
    yield CV(q0, q2)


def translate_ccxpow_to_cnotpow(gate: CCXPow) -> Iterator[Union[CNot, CNotPow]]:
    """Decomposition of powers of CCNot gates to powers of CNot gates."""
    q0, q1, q2 = gate.qubits
    (t,) = gate.params
    yield CNotPow(t / 2, q1, q2)
    yield CNot(q0, q1)
    yield CNotPow(-t / 2, q1, q2)
    yield CNot(q0, q1)
    yield CNotPow(t / 2, q0, q2)


def translate_ccz_to_adjacent_cnot(gate: CCZ) -> Iterator[Union[T, CNot, T_H]]:
    """Decomposition of CCZ gate into 8 CNot gates.
    Respects linear adjacency of qubits.
    ::

        ───●───     ───T───●────────────●───────●────────●────────
           │               │            │       │        │
        ───●───  =  ───T───X───●───T⁺───X───●───X────●───X────●───
           │                   │            │        │        │
        ───●───     ───T───────X───T────────X───T⁺───X───T⁺───X───
    """
    # Kudos: adapted from Cirq

    q0, q1, q2 = gate.qubits

    yield T(q0)
    yield T(q1)
    yield T(q2)

    yield CNot(q0, q1)
    yield CNot(q1, q2)

    yield T_H(q1)
    yield T(q2)

    yield CNot(q0, q1)
    yield CNot(q1, q2)

    yield T_H(q2)

    yield CNot(q0, q1)
    yield CNot(q1, q2)

    yield T_H(q2)

    yield CNot(q0, q1)
    yield CNot(q1, q2)


def translate_ccz_to_ccnot(gate: CCZ) -> Iterator[Union[H, CCNot]]:
    """Convert  CCZ gate to CCNot gate using Hadamards
    ::

        ───●───     ───────●───────
           │               │
        ───●───  =  ───────●───────
           │               │
        ───X───     ───H───●───H───

    """
    q0, q1, q2 = gate.qubits

    yield H(q2)
    yield CCNot(q0, q1, q2)
    yield H(q2)


def translate_cswap_to_ccnot(gate: CSwap) -> Iterator[Union[CNot, CCNot]]:
    """Convert a CSwap gate to a circuit with a CCNot and 2 CNots"""
    q0, q1, q2 = gate.qubits
    yield CNot(q2, q1)
    yield CCNot(q0, q1, q2)
    yield CNot(q2, q1)


def translate_cswap_to_cnot(
    gate: CSwap,
) -> Iterator[Union[CNot, H, T, T_H, V, V_H, S, S_H]]:
    """Adjacency respecting decomposition of CSwap to CNot.
    Assumes that q0 (control) is next to q1 is next to q2.
    ::

        ───●───     ───T───────────●────────────●───────●────────────────●───────────────────────
           │                       │            │       │                │
        ───x───  =  ───X───T───────X───●───T⁺───X───●───X────●───X^1/2───X───●───X^1/2───────────
           │           │               │            │        │               │
        ───x───     ───●───Y^3/2───T───X───T────────X───T⁺───X───T⁺──────────X───S───────X^3/2───
    """  # noqa: W291, E501
    # Kudos: Adapted from Cirq
    q0, q1, q2 = gate.qubits

    yield CNot(q2, q1)
    yield H(q2)
    yield S(q2).H
    yield T(q0)
    yield T(q1)
    yield T(q2).H
    yield CNot(q0, q1)
    yield CNot(q1, q2)
    yield T(q1).H
    yield T(q2)
    yield CNot(q0, q1)
    yield CNot(q1, q2)
    yield T(q2).H
    yield CNot(q0, q1)
    yield CNot(q1, q2)
    yield T(q2).H
    yield V(q1)
    yield CNot(q0, q1)
    yield CNot(q1, q2)
    yield S(q2)
    yield V(q1)
    yield V(q2).H


def translate_cswap_inside_to_cnot(
    gate: CSwap,
) -> Iterator[Union[CNot, H, T, T_H, V, V_H, S, S_H]]:
    """Adjacency respecting decomposition of centrally controlled CSwap to CNot.
    Assumes that q0 (control) is next to q1 is next to q2.
    ::

        ───x───      ───●───X───T────────────────────●────────────●────────────X───────●───V⁺────────────
           │            │   │                        │            │            │       │
        ───●───  =   ───X───●───X───────────●───T⁺───X───●───T────X───●───V────●───●───X───●─────────────
           │                    │           │            │            │            │       │
        ───x───      ───────────●───H───T───X───T────────X───T⁺───────X───T⁺───────X───────X────H───S⁺───
    """  # noqa: W291, E501
    # Kudos: Adapted from Cirq
    q0, q1, q2 = gate.qubits

    yield CNot(q1, q0)
    yield CNot(q0, q1)
    yield CNot(q2, q0)
    yield H(q2)
    yield T(q2)
    yield CNot(q0, q2)
    yield T(q1)
    yield T_H(q0)
    yield T(q2)
    yield CNot(q1, q0)
    yield CNot(q0, q2)
    yield T(q0)
    yield T_H(q2)
    yield CNot(q1, q0)
    yield CNot(q0, q2)
    yield V(q0)
    yield T_H(q2)
    yield CNot(q0, q1)
    yield CNot(q0, q2)
    yield CNot(q1, q0)
    yield CNot(q0, q2)
    yield H(q2)
    yield S_H(q2)
    yield V_H(q1)


def translate_deutsch_to_barenco(gate: Deutsch) -> Iterator[Barenco]:
    """Translate a 3-qubit Deutsch gate to five 2-qubit Barenco gates.

    Ref:
        A Universal Two–Bit Gate for Quantum Computation, A. Barenco (1995)
        https://arxiv.org/pdf/quant-ph/9505016.pdf  :cite:`Barenco1995a`
    """
    q0, q1, q2 = gate.qubits
    (theta,) = gate.params
    yield Barenco(0, var.PI / 4, theta / 2, q1, q2)
    yield Barenco(0, var.PI / 4, theta / 2, q0, q2)
    yield Barenco(0, var.PI / 2, var.PI / 2, q0, q1)
    yield Barenco(var.PI, -var.PI / 4, theta / 2, q1, q2)
    yield Barenco(0, var.PI / 2, var.PI / 2, q0, q1)


TRANSLATORS: Dict[str, Callable] = {}
TRANSLATORS = {
    name: func for name, func in globals().items() if name.startswith("translate_")
}


# Note: Translators auto-magically added
__all__ = ("circuit_translate", "select_translators", "TRANSLATORS") + tuple(
    TRANSLATORS.keys()
)

# fin
